from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError, UserResourcePolicyNotFound
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.resource.types import UserResourcePolicyData
from ai.backend.manager.models.resource_policy import UserResourcePolicyRow
from ai.backend.manager.repositories.base.creator import Creator, execute_creator
from ai.backend.manager.repositories.base.updater import Updater, execute_updater

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

user_resource_policy_db_source_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.DB_SOURCE, layer=LayerType.USER_RESOURCE_POLICY_DB_SOURCE)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=5,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class UserResourcePolicyDBSource:
    """
    Database source for user resource policy operations.
    Handles all database operations for user resource policies.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @user_resource_policy_db_source_resilience.apply()
    async def create(self, creator: Creator[UserResourcePolicyRow]) -> UserResourcePolicyData:
        """Creates a new user resource policy."""
        async with self._db.begin_session() as db_sess:
            result = await execute_creator(db_sess, creator)
            return result.row.to_dataclass()

    @user_resource_policy_db_source_resilience.apply()
    async def get_by_name(self, name: str) -> UserResourcePolicyData:
        """Retrieves a user resource policy by name."""
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            query = sa.select(UserResourcePolicyRow).where(UserResourcePolicyRow.name == name)
            row = await db_sess.scalar(query)
            if row is None:
                raise UserResourcePolicyNotFound(
                    f"User resource policy with name {name} not found."
                )
            return row.to_dataclass()

    @user_resource_policy_db_source_resilience.apply()
    async def update(self, updater: Updater[UserResourcePolicyRow]) -> UserResourcePolicyData:
        """Updates an existing user resource policy."""
        async with self._db.begin_session() as db_sess:
            result = await execute_updater(db_sess, updater)
            if result is None:
                raise UserResourcePolicyNotFound(
                    f"User resource policy with name {updater.pk_value} not found."
                )
            return result.row.to_dataclass()

    @user_resource_policy_db_source_resilience.apply()
    async def delete(self, name: str) -> UserResourcePolicyData:
        """Deletes a user resource policy."""
        async with self._db.begin_session() as db_sess:
            delete_stmt = (
                sa.delete(UserResourcePolicyRow)
                .where(UserResourcePolicyRow.name == name)
                .returning(UserResourcePolicyRow)
            )
            query_stms = (
                sa.select(UserResourcePolicyRow)
                .from_statement(delete_stmt)
                .execution_options(populate_existing=True)
            )
            row: UserResourcePolicyRow | None = await db_sess.scalar(query_stms)
            if row is None:
                raise UserResourcePolicyNotFound(
                    f"User resource policy with name {name} not found."
                )
            await db_sess.delete(row)
            return row.to_dataclass()



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.artifact.types import ArtifactRevisionData
from ai.backend.manager.repositories.artifact.options import ArtifactRevisionConditions
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.services.artifact_revision.actions.search import (
    SearchArtifactRevisionsAction,
)
from ai.backend.manager.services.artifact_revision.processors import ArtifactRevisionProcessors


async def load_artifact_revisions_by_ids(
    processor: ArtifactRevisionProcessors,
    revision_ids: Sequence[uuid.UUID],
) -> list[ArtifactRevisionData | None]:
    """Batch load artifact revisions by their IDs.

    Args:
        processor: The artifact revision processor.
        revision_ids: Sequence of revision UUIDs to load.

    Returns:
        List of ArtifactRevisionData (or None if not found) in the same order as revision_ids.
    """
    if not revision_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(revision_ids)),
        conditions=[ArtifactRevisionConditions.by_ids(revision_ids)],
    )

    action_result = await processor.search_revision.wait_for_complete(
        SearchArtifactRevisionsAction(querier=querier)
    )

    revision_map = {revision.id: revision for revision in action_result.data}
    return [revision_map.get(revision_id) for revision_id in revision_ids]



"""Export report registry base class."""

from __future__ import annotations

from ai.backend.manager.repositories.base.export import ReportDef
from ai.backend.manager.repositories.export.reports.audit_log import AUDIT_LOG_REPORT
from ai.backend.manager.repositories.export.reports.keypair import KEYPAIR_REPORT
from ai.backend.manager.repositories.export.reports.project import PROJECT_REPORT
from ai.backend.manager.repositories.export.reports.session import SESSION_REPORT
from ai.backend.manager.repositories.export.reports.user import USER_REPORT


class ExportReportRegistry:
    """Registry for export report definitions.

    Manages a collection of ReportDef instances that can be
    looked up by their report_key.
    """

    def __init__(self) -> None:
        """Initialize an empty registry."""
        self._reports: dict[str, ReportDef] = {}

    def register(self, report: ReportDef) -> None:
        """Register a report definition.

        Args:
            report: The ReportDef to register
        """
        self._reports[report.report_key] = report

    def get(self, report_key: str) -> ReportDef | None:
        """Get ReportDef by report_key.

        Args:
            report_key: The unique identifier for the report

        Returns:
            ReportDef if found, None otherwise
        """
        return self._reports.get(report_key)

    def list_all(self) -> list[ReportDef]:
        """Return all registered reports.

        Returns:
            List of all ReportDef instances
        """
        return list(self._reports.values())

    @classmethod
    def create_default(cls) -> ExportReportRegistry:
        """Create the default export report registry with built-in reports.

        Returns:
            ExportReportRegistry with all built-in reports registered
        """
        registry = cls()
        registry.register(AUDIT_LOG_REPORT)
        registry.register(KEYPAIR_REPORT)
        registry.register(PROJECT_REPORT)
        registry.register(SESSION_REPORT)
        registry.register(USER_REPORT)
        return registry



"""Generic recorder module for sokovan coordinators.

This module provides base recorder functionality that can be specialized
for different coordinator contexts (scheduler, deployment, route).
"""

from .context import RecorderContext
from .exceptions import (
    NestedPhaseError,
    RecorderError,
    StepWithoutPhaseError,
)
from .pool import RecordPool
from .recorder import TransitionRecorder
from .types import (
    EntityIdT,
    ExecutionRecord,
    PhaseRecord,
    StepRecord,
    StepStatus,
)

__all__ = [
    "EntityIdT",
    "ExecutionRecord",
    "NestedPhaseError",
    "PhaseRecord",
    "RecordPool",
    "RecorderContext",
    "RecorderError",
    "StepRecord",
    "StepStatus",
    "StepWithoutPhaseError",
    "TransitionRecorder",
]



from __future__ import annotations

from collections.abc import Iterable, Mapping
from datetime import datetime
from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Any, Self

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID
from strawberry.scalars import JSON

from ai.backend.common.types import AgentId
from ai.backend.manager.api.gql.base import OrderDirection, StringFilter
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext

if TYPE_CHECKING:
    from ai.backend.manager.api.gql.kernel.types import (
        KernelV2ConnectionGQL,
        KernelV2FilterGQL,
        KernelV2OrderByGQL,
    )
from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.agent.types import AgentDetailData, AgentStatus
from ai.backend.manager.models.rbac.permission_defs import AgentPermission
from ai.backend.manager.repositories.agent.options import AgentConditions, AgentOrders
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.scheduler.options import (
    KernelConditions,
)


@strawberry.enum(
    name="AgentPermission",
    description="Added in 26.1.0. Permissions related to agent operations",
)
class AgentPermissionGQL(StrEnum):
    READ_ATTRIBUTE = "read_attribute"
    UPDATE_ATTRIBUTE = "update_attribute"
    CREATE_COMPUTE_SESSION = "create_compute_session"
    CREATE_SERVICE = "create_service"

    @classmethod
    def from_agent_permission(cls, permission: AgentPermission) -> AgentPermissionGQL:
        match permission:
            case AgentPermission.READ_ATTRIBUTE:
                return AgentPermissionGQL.READ_ATTRIBUTE
            case AgentPermission.UPDATE_ATTRIBUTE:
                return AgentPermissionGQL.UPDATE_ATTRIBUTE
            case AgentPermission.CREATE_COMPUTE_SESSION:
                return AgentPermissionGQL.CREATE_COMPUTE_SESSION
            case AgentPermission.CREATE_SERVICE:
                return AgentPermissionGQL.CREATE_SERVICE


@strawberry.enum(
    name="AgentOrderField",
    description="Added in 26.1.0. Order by specification for agents",
)
class AgentOrderFieldGQL(StrEnum):
    ID = "id"
    STATUS = "status"
    FIRST_CONTACT = "first_contact"
    SCALING_GROUP = "scaling_group"
    SCHEDULABLE = "schedulable"


@strawberry.input(
    name="AgentStatusFilter",
    description=dedent_strip("""
        Added in 26.1.0. Filter options for agent status within AgentFilter.
        It includes options to filter whether agent status is in a specific list or equals a specific value.
    """),
)
class AgentStatusFilterGQL:
    in_: list[AgentStatus] | None = strawberry.field(name="in", default=None)
    equals: AgentStatus | None = None


@strawberry.input(
    name="AgentFilter", description="Added in 26.1.0. Filter options for querying agents"
)
class AgentFilterGQL(GQLFilter):
    id: StringFilter | None = None
    status: AgentStatusFilterGQL | None = None
    schedulable: bool | None = None
    scaling_group: StringFilter | None = None

    AND: list[AgentFilterGQL] | None = None
    OR: list[AgentFilterGQL] | None = None
    NOT: list[AgentFilterGQL] | None = None

    def build_conditions(self) -> list[QueryCondition]:
        field_conditions: list[QueryCondition] = []
        if self.id is not None:
            name_condition = self.id.build_query_condition(
                contains_factory=AgentConditions.by_id_contains,
                equals_factory=AgentConditions.by_id_equals,
                starts_with_factory=AgentConditions.by_id_starts_with,
                ends_with_factory=AgentConditions.by_id_ends_with,
            )
            if name_condition is not None:
                field_conditions.append(name_condition)
        if self.status is not None:
            if self.status.in_ is not None:
                field_conditions.append(AgentConditions.by_status_contains(self.status.in_))
            if self.status.equals is not None:
                field_conditions.append(AgentConditions.by_status_equals(self.status.equals))
        if self.schedulable is not None:
            field_conditions.append(AgentConditions.by_schedulable(self.schedulable))
        if self.scaling_group is not None:
            scaling_group_condition = self.scaling_group.build_query_condition(
                contains_factory=AgentConditions.by_scaling_group_contains,
                equals_factory=AgentConditions.by_scaling_group_equals,
                starts_with_factory=AgentConditions.by_scaling_group_starts_with,
                ends_with_factory=AgentConditions.by_scaling_group_ends_with,
            )
            if scaling_group_condition is not None:
                field_conditions.append(scaling_group_condition)

        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(name="AgentOrderBy", description="Added in 26.1.0. Options for ordering agents")
class AgentOrderByGQL(GQLOrderBy):
    field: AgentOrderFieldGQL
    direction: OrderDirection = OrderDirection.ASC

    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case AgentOrderFieldGQL.ID:
                return AgentOrders.id(ascending)
            case AgentOrderFieldGQL.STATUS:
                return AgentOrders.status(ascending)
            case AgentOrderFieldGQL.FIRST_CONTACT:
                return AgentOrders.first_contact(ascending)
            case AgentOrderFieldGQL.SCALING_GROUP:
                return AgentOrders.scaling_group(ascending)
            case AgentOrderFieldGQL.SCHEDULABLE:
                return AgentOrders.schedulable(ascending)


@strawberry.type(name="AgentResource", description="Added in 25.15.0")
class AgentResourceGQL:
    capacity: JSON = strawberry.field(
        description=dedent_strip("""
            Total hardware resource capacity available on the agent.
            Expressed as a JSON object containing resource slots (e.g., cpu, mem, accelerators).
            Each slot represents the maximum amount of that resource type the agent can provide.
        """)
    )
    used: JSON = strawberry.field(
        description=dedent_strip("""
            Total amount of resources currently consumed by running and scheduled compute sessions.
            Includes both the requested resources for sessions being prepared and already allocated
            resources for active sessions. The sum of occupied resources across all session states
            that occupy agent resources (PREPARING, PULLING, RUNNING, RESTARTING, etc.).
            Expressed as a JSON object with the same structure as capacity.
        """)
    )
    free: JSON = strawberry.field(
        description=dedent_strip("""
            Available resources for scheduling new compute sessions (capacity - used).
            This represents the maximum resources that can be allocated to new sessions
            without exceeding the agent's capacity. Expressed as a JSON object with
            the same structure as capacity.
        """)
    )


@strawberry.type(name="AgentStats", description="Added in 25.15.0")
class AgentStatsGQL:
    total_resource: AgentResourceGQL = strawberry.field(description="Added in 25.15.0")


@strawberry.type(
    name="AgentStatusInfo",
    description="Added in 26.1.0. Status and lifecycle information for an agent",
)
class AgentStatusInfoGQL:
    status: AgentStatus = strawberry.field(
        description=dedent_strip("""
            Current operational status of the agent.
            Indicates whether the agent is ALIVE (active and reachable), LOST (unreachable),
            TERMINATED (intentionally stopped), or RESTARTING (in recovery process).
        """)
    )
    status_changed: datetime | None = strawberry.field(
        description=dedent_strip("""
            Timestamp when the agent last changed its status.
            Updated whenever the agent transitions between different status states
            (e.g., from ALIVE to LOST, or RESTARTING to ALIVE).
            Will be null if the agent status has never changed since initial registration.
        """)
    )
    first_contact: datetime | None = strawberry.field(
        description=dedent_strip("""
            Timestamp when the agent first registered with the manager.
            This value remains constant throughout the agent's lifecycle and can be used
            to track the agent's age or identify when it was initially deployed.
        """)
    )
    lost_at: datetime | None = strawberry.field(
        description=dedent_strip("""
            Timestamp when the agent was marked as lost or unreachable.
            Set when the manager detects the agent has stopped sending heartbeats.
            Will be null if the agent has never been lost or is currently alive.
        """)
    )
    schedulable: bool = strawberry.field(
        description=dedent_strip("""
            Indicates whether the agent is available for scheduling new compute sessions.
            An agent can be non-schedulable due to maintenance mode, resource constraints or other operational reasons by admin.
            When false, no new sessions will be assigned to this agent.
        """)
    )


@strawberry.type(
    name="ComputePluginEntry",
    description=(
        "Added in 26.1.0. A single compute plugin entry representing one plugin and its metadata."
    ),
)
class ComputePluginEntryGQL:
    """Single compute plugin entry with plugin name and metadata."""

    plugin_name: str = strawberry.field(
        description=(
            "Name of the compute plugin (e.g., 'cuda', 'rocm', 'tpu'). "
            "This identifier corresponds to the accelerator or resource type supported by the agent."
        )
    )
    value: str = strawberry.field(
        description=(
            "Plugin value string containing plugin-specific information. "
            "The content varies by plugin type and may include version or configuration details."
        )
    )


@strawberry.type(
    name="ComputePlugins",
    description=(
        "Added in 26.1.0. A collection of compute plugins available on an agent. "
        "Each entry specifies a plugin name and its associated metadata."
    ),
)
class ComputePluginsGQL:
    """Compute plugins container with multiple plugin entries."""

    entries: list[ComputePluginEntryGQL] = strawberry.field(
        description=(
            "List of compute plugins. Each entry contains a plugin name and its metadata. "
            "The list includes all accelerator and resource type plugins installed on the agent."
        )
    )

    @classmethod
    def from_mapping(cls, plugins: Mapping[str, str]) -> ComputePluginsGQL:
        """Convert a mapping of plugin name to value to GraphQL type."""
        entries = [
            ComputePluginEntryGQL(plugin_name=name, value=value) for name, value in plugins.items()
        ]
        return cls(entries=entries)


@strawberry.type(
    name="AgentSystemInfo",
    description="Added in 26.1.0. System and configuration information for an agent",
)
class AgentSystemInfoGQL:
    architecture: str = strawberry.field(
        description=dedent_strip("""
            Hardware architecture of the agent's host system (e.g., "x86_64", "aarch64").
            Used to match compute sessions with compatible container images and ensure
            proper binary execution on the agent.
        """)
    )
    version: str = strawberry.field(
        description=dedent_strip("""
            Version string of the Backend.AI agent software running on this node.
            Follows semantic versioning (e.g., "26.1.0") and helps identify
            compatibility and available features.
        """)
    )
    auto_terminate_abusing_kernel: bool = strawberry.field(
        description=dedent_strip("""
            Legacy configuration flag, no longer actively used in the system.
            Retained for backward compatibility and schema consistency.
            Originally intended to control automatic termination of misbehaving sessions.
        """),
        deprecation_reason="Legacy feature no longer in use.",
    )
    compute_plugins: ComputePluginsGQL = strawberry.field(
        description=dedent_strip("""
            List of compute plugin metadata supported by this agent.
            Each plugin represents a specific accelerator or resource type (e.g., CUDA).
            Entries contain plugin names and their associated metadata with
            plugin-specific configuration and capabilities.
        """)
    )


@strawberry.type(
    name="AgentNetworkInfo",
    description="Added in 26.1.0. Network-related information for an agent",
)
class AgentNetworkInfoGQL:
    region: str = strawberry.field(description="Logical region where the agent is deployed.")
    addr: str = strawberry.field(
        description=dedent_strip("""
            Network address and port where the agent can be reached (format: "host:port").
            This is the bind or advertised address used by the manager to communicate
            with the agent for session lifecycle management and health monitoring.
        """)
    )


@strawberry.type(
    name="AgentV2", description="Added in 26.1.0. Strawberry-based Agent type replacing AgentNode."
)
class AgentV2GQL(Node):
    _agent_id: strawberry.Private[AgentId]
    id: NodeID[str]
    resource_info: AgentResourceGQL = strawberry.field(
        description=dedent_strip("""
            Hardware resource capacity, usage, and availability information.
            Contains capacity (total), used (occupied by sessions), and free (available) resource slots
            including CPU cores, memory, accelerators (GPUs, TPUs), and other compute resources.
        """)
    )
    status_info: AgentStatusInfoGQL = strawberry.field(
        description=dedent_strip("""
            Current operational status and lifecycle timestamps.
            Includes the agent's status (ALIVE, LOST, etc.), status change history,
            initial registration time, and schedulability state.
        """)
    )
    system_info: AgentSystemInfoGQL = strawberry.field(
        description=dedent_strip("""
            System configuration and software version information.
            Contains the host architecture, agent software version, and available compute plugins
            for accelerators and specialized hardware.
        """)
    )
    network_info: AgentNetworkInfoGQL = strawberry.field(
        description=dedent_strip("""
            Network location and connectivity information.
            Provides the agent's region and network address for manager-to-agent communication.
        """)
    )
    permissions: list[AgentPermissionGQL] = strawberry.field(
        description=dedent_strip("""
            List of permissions the current authenticated user has on this agent.
            Determines which operations (read attributes, create sessions, etc.)
            the user can perform on this specific agent based on RBAC policies.
        """)
    )
    scaling_group: str = strawberry.field(
        description=dedent_strip("""
            Name of the scaling group this agent belongs to.
            Scaling groups are logical collections of agents used for resource scheduling,
            quota management, and workload isolation across different user groups or projects.
        """)
    )

    @strawberry.field(description="Added in 26.1.0. Load the container count for this agent.")  # type: ignore[misc]
    async def container_count(
        self,
        info: Info[StrawberryGQLContext],
    ) -> int:
        """
        Get the container count for a specific agent.
        """
        return await info.context.data_loaders.container_count_loader.load(self._agent_id)

    @strawberry.field(  # type: ignore[misc]
        description="Added in 26.2.0. List of kernels running on this agent with pagination support."
    )
    async def kernels(
        self,
        info: Info[StrawberryGQLContext],
        filter: Annotated[
            KernelV2FilterGQL, strawberry.lazy("ai.backend.manager.api.gql.kernel.types")
        ]
        | None = None,
        order_by: list[
            Annotated[
                KernelV2OrderByGQL, strawberry.lazy("ai.backend.manager.api.gql.kernel.types")
            ]
        ]
        | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> Annotated[
        KernelV2ConnectionGQL, strawberry.lazy("ai.backend.manager.api.gql.kernel.types")
    ]:
        """Fetch kernels associated with this agent."""
        from ai.backend.manager.api.gql.kernel.fetcher import fetch_kernels

        return await fetch_kernels(
            info=info,
            filter=filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
            base_conditions=[KernelConditions.by_agent_id(str(self._agent_id))],
        )

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.agent_loader.load_many([
            AgentId(nid) for nid in node_ids
        ])
        return [cls.from_agent_detail_data(data) if data is not None else None for data in results]

    @classmethod
    def from_agent_detail_data(cls, detail_data: AgentDetailData) -> Self:
        data = detail_data.agent

        return cls(
            _agent_id=data.id,
            id=ID(data.id),
            resource_info=AgentResourceGQL(
                capacity=data.available_slots.to_json(),
                used=data.actual_occupied_slots.to_json(),
                free=(data.available_slots - data.actual_occupied_slots).to_json(),
            ),
            status_info=AgentStatusInfoGQL(
                status=data.status,
                status_changed=data.status_changed,
                first_contact=data.first_contact,
                lost_at=data.lost_at,
                schedulable=data.schedulable,
            ),
            system_info=AgentSystemInfoGQL(
                architecture=data.architecture,
                version=data.version,
                auto_terminate_abusing_kernel=data.auto_terminate_abusing_kernel,
                compute_plugins=ComputePluginsGQL.from_mapping(data.compute_plugins),
            ),
            network_info=AgentNetworkInfoGQL(
                region=data.region,
                addr=data.addr,
            ),
            permissions=[
                AgentPermissionGQL.from_agent_permission(p) for p in detail_data.permissions
            ],
            scaling_group=data.scaling_group,
        )


AgentV2Edge = Edge[AgentV2GQL]


@strawberry.type(
    description="Added in 26.1.0. Relay-style connection type for paginated lists of agents"
)
class AgentV2Connection(Connection[AgentV2GQL]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count



from __future__ import annotations

import uuid
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Self, cast

import graphene
import sqlalchemy as sa
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.exc import NoResultFound

from ai.backend.manager.data.deployment.types import RouteStatus
from ai.backend.manager.errors.service import RoutingNotFound
from ai.backend.manager.models.routing import RoutingRow

from .base import InferenceSessionError, Item, PaginatedList

if TYPE_CHECKING:
    from ai.backend.manager.models.endpoint import EndpointRow

    from .schema import GraphQueryContext

__all__ = ("Routing", "RoutingList")


class Routing(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (Item,)

    routing_id = graphene.UUID()
    endpoint = graphene.String()
    session = graphene.UUID()
    status = graphene.String()
    traffic_ratio = graphene.Float()
    created_at = GQLDateTime()
    error = InferenceSessionError()
    error_data = graphene.JSONString()

    live_stat = graphene.JSONString(description="Added in 24.12.0.")

    _endpoint_row: EndpointRow

    @classmethod
    def from_dto(cls, dto: Any) -> Self | None:
        if dto is None:
            return None
        return cls(
            routing_id=dto.id,
            endpoint=dto.endpoint,
            session=dto.session,
            status=dto.status.name,
            traffic_ratio=dto.traffic_ratio,
            created_at=dto.created_at,
            error_data=dto.error_data,
        )

    @classmethod
    async def from_row(
        cls,
        ctx: GraphQueryContext,
        row: RoutingRow,
        endpoint: EndpointRow | None = None,
    ) -> Routing:
        ret = cls(
            routing_id=row.id,
            endpoint=(endpoint or row.endpoint_row).url,
            session=row.session,
            status=row.status.name,
            traffic_ratio=row.traffic_ratio,
            created_at=row.created_at,
            error_data=row.error_data,
        )
        ret._endpoint_row = endpoint or row.endpoint_row
        return ret

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        *,
        endpoint_id: uuid.UUID | None = None,
        project: uuid.UUID | None = None,
        domain_name: str | None = None,
        user_uuid: uuid.UUID | None = None,
    ) -> int:
        query = sa.select(sa.func.count()).select_from()
        if endpoint_id is not None:
            query = query.where(RoutingRow.endpoint == endpoint_id)
        if project:
            query = query.filter(RoutingRow.project == project)
        if domain_name:
            query = query.filter(RoutingRow.domain == domain_name)
        if user_uuid:
            query = query.filter(RoutingRow.session_owner == user_uuid)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar() or 0

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        endpoint_id: uuid.UUID | None = None,
        filter: str | None = None,
        order: str | None = None,
        project: uuid.UUID | None = None,
        domain_name: str | None = None,
        user_uuid: uuid.UUID | None = None,
    ) -> Sequence[Routing]:
        query = (
            sa.select(RoutingRow)
            .limit(limit)
            .offset(offset)
            .order_by(sa.desc(RoutingRow.created_at))
        )

        if endpoint_id is not None:
            query = query.where(RoutingRow.endpoint == endpoint_id)
        if project:
            query = query.filter(RoutingRow.project == project)
        if domain_name:
            query = query.filter(RoutingRow.domain == domain_name)
        if user_uuid:
            query = query.filter(RoutingRow.session_owner == user_uuid)
        """
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            parser = QueryOrderParser(cls._queryorder_colmap)
            query = parser.append_ordering(query, order)
        """
        async with ctx.db.begin_readonly_session() as session:
            return [
                await cls.from_row(ctx, row) async for row in (await session.stream_scalars(query))
            ]

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        endpoint_id: uuid.UUID,
        *,
        project: uuid.UUID | None = None,
        domain_name: str | None = None,
        user_uuid: uuid.UUID | None = None,
    ) -> Sequence[Routing]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await RoutingRow.list(
                session,
                endpoint_id,
                project=project,
                domain=domain_name,
                user_uuid=user_uuid,
            )
        return [await Routing.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx: GraphQueryContext,
        *,
        routing_id: uuid.UUID,
        project: uuid.UUID | None = None,
        domain_name: str | None = None,
        user_uuid: uuid.UUID | None = None,
    ) -> Routing:
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await RoutingRow.get(
                    session, routing_id, project=project, domain=domain_name, user_uuid=user_uuid
                )
        except NoResultFound as e:
            raise RoutingNotFound from e
        return await Routing.from_row(ctx, row)

    async def resolve_error(self, info: graphene.ResolveInfo) -> Any:
        if self.status != RouteStatus.FAILED_TO_START or not self.error_data:
            return None
        match self.error_data["type"]:
            case "session_cancelled":
                session_id = self.error_data["session_id"]
            case _:
                session_id = None
        return InferenceSessionError(
            session_id=session_id,
            errors=[
                InferenceSessionError.InferenceSessionErrorInfo(
                    src=e["src"], name=e["name"], repr=e["repr"]
                )
                for e in self.error_data["errors"]
            ],
        )

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Mapping[str, Any] | None:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(graph_ctx, "EndpointStatistics.by_replica")
        return cast(
            Mapping[str, Any] | None, await loader.load((self._endpoint_row.id, self.routing_id))
        )


class RoutingList(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(Routing, required=True)



from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

metric_repository_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.METRIC_REPOSITORY)),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class MetricRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import click

from ai.backend.cli.params import BoolExprType, OptionalType
from ai.backend.common.cli import LazyGroup

from .context import CLIContext

if TYPE_CHECKING:
    from ai.backend.logging import BraceStyleAdapter

# LogLevel values for click.Choice - avoid importing ai.backend.logging at module level
_LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE", "NOTSET"]


def _get_logger() -> BraceStyleAdapter:
    import logging

    from ai.backend.logging import BraceStyleAdapter

    return BraceStyleAdapter(logging.getLogger("ai.backend.manager.cli"))


@click.group(invoke_without_command=False, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=True,
        path_type=pathlib.Path,
    ),
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice(_LOG_LEVELS, case_sensitive=False),
    default="NOTSET",
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    log_level: str,
    debug: bool,
    config_path: pathlib.Path | None = None,
) -> None:
    """
    Manager Administration CLI
    """
    from setproctitle import setproctitle

    from ai.backend.logging.types import LogLevel

    setproctitle("backend.ai: manager.cli")
    if debug:
        log_level = "DEBUG"

    ctx.obj = ctx.with_resource(CLIContext(config_path=config_path, log_level=LogLevel(log_level)))


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--psql-container",
    "container_name",
    type=str,
    default=None,
    metavar="ID_OR_NAME",
    help=(
        "Open a postgres client shell using the psql executable "
        "shipped with the given postgres container. "
        'If not set or set as an empty string "", it will auto-detect '
        "the psql container from the halfstack. "
        'If set "-", it will use the host-provided psql executable. '
        "You may append additional arguments passed to the psql cli command. "
        "[default: auto-detect from halfstack]"
    ),
)
@click.option(
    "--psql-help",
    is_flag=True,
    help="Show the help text of the psql command instead of this dbshell command.",
)
@click.argument("psql_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def dbshell(
    cli_ctx: CLIContext, container_name: str | None, psql_help: bool, psql_args: tuple[str, ...]
) -> None:
    """
    Run the database shell.

    All additional arguments and options except `--psql-container` and `--psql-help` are
    transparently forwarded to the psql command.
    For instance, you can use `-c` to execute a psql/SQL statement on the command line.

    Note that you do not have to specify connection-related options
    because the dbshell command fills out them from the manager configuration.
    """
    import asyncio
    import subprocess
    import sys

    from ai.backend.cli.types import ExitCode

    bootstrap_config = asyncio.run(cli_ctx.get_bootstrap_config())
    db_config = bootstrap_config.db
    _psql_args: list[str]
    if psql_help:
        _psql_args = ["--help"]
    else:
        _psql_args = list(psql_args)
    if not container_name:
        # Try to get the database container name of the halfstack
        candidate_container_names = subprocess.check_output(
            ["docker", "ps", "--format", "{{.Names}}", "--filter", "name=half-db"],
        )
        if not candidate_container_names:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
        container_name = None
        name_list = candidate_container_names.decode("utf-8").splitlines()
        for name in name_list:
            if "exporter" in name:
                continue  # Skip exporter containers
            container_name = name
            break
        if not container_name:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
    elif container_name == "-":
        # Use the host-provided psql command
        cmd = [
            "psql",
            (f"postgres://{db_config.user}:{db_config.password}@{db_config.addr}/{db_config.name}"),
            *_psql_args,
        ]
        subprocess.run(cmd)
        return
    # Use the container to start the psql client command
    _get_logger().info(f"using the db container {container_name} ...")
    cmd = [
        "docker",
        "exec",
        "-i",
        "-t",
        container_name,
        "psql",
        "-U",
        db_config.user,
        "-d",
        db_config.name,
        *_psql_args,
    ]
    subprocess.run(cmd)


@main.command()
@click.pass_obj
def generate_api_keypair(_cli_ctx: CLIContext) -> None:
    """
    Generate a manager API keypair and print it out to stdout.
    """
    from ai.backend.manager.models.keypair import generate_keypair as _gen_keypair

    _get_logger().info("Generating a manager API keypair...")
    ak, sk = _gen_keypair()
    print(f"Access Key: {ak} ({len(ak)} bytes)")
    print(f"Secret Key: {sk} ({len(sk)} bytes)")


@main.command()
@click.argument(
    "dst_dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        exists=True,
        writable=True,
        path_type=pathlib.Path,
    ),
)
@click.argument(
    "name",
    type=str,
)
@click.pass_obj
def generate_rpc_keypair(_cli_ctx: CLIContext, dst_dir: pathlib.Path, name: str) -> None:
    """
    Generate a ZeroMQ CURVE keypair for use in manager-agent RPC and print it out to stdout.

    \b
    DST_DIR: The target directory to store .key/.key_secret files
    NAME: The name of generated key files
    """
    from zmq.auth.certs import create_certificates, load_certificate

    from ai.backend.manager.errors.resource import ConfigurationLoadFailed

    _get_logger().info("Generating a RPC keypair...")
    public_key_path, secret_key_path = create_certificates(dst_dir, name)
    public_key, secret_key = load_certificate(secret_key_path)
    if secret_key is None:
        raise ConfigurationLoadFailed("Failed to load secret key from certificate")
    print(f"Public Key: {public_key.decode('ascii')} (stored at {public_key_path})")
    print(f"Secret Key: {secret_key.decode('ascii')} (stored at {secret_key_path})")


@main.command()
@click.option(
    "-r",
    "--retention",
    type=str,
    default="1yr",
    help="The retention limit. e.g., 20d, 1mo, 6mo, 1yr",
)
@click.option(
    "-v",
    "--vacuum-full",
    type=OptionalType(BoolExprType),
    default=False,
    help=(
        "Reclaim storage occupied by dead tuples."
        "If not set or set False, it will run VACUUM without FULL. "
        "If set True, it will run VACUUM FULL. "
        "When VACUUM FULL is being processed, the database is locked. "
        "[default: False]"
    ),
)
@click.pass_obj
def clear_history(cli_ctx: CLIContext, retention: str, vacuum_full: bool) -> None:
    """
    Delete old records from the kernels, error_logs tables and
    invoke the PostgreSQL's vaccuum operation to clear up the actual disk space.
    """
    import asyncio
    import uuid
    from datetime import UTC, datetime
    from typing import cast

    import sqlalchemy as sa
    from more_itertools import chunked

    from ai.backend.common.validators import TimeDuration
    from ai.backend.manager.models.error_logs import error_logs
    from ai.backend.manager.models.kernel import kernels
    from ai.backend.manager.models.session import SessionRow
    from ai.backend.manager.models.utils import connect_database, vacuum_db

    from .context import redis_ctx

    log = _get_logger()
    today = datetime.now(UTC)
    duration = TimeDuration()
    expiration_date = today - duration.check_and_return(retention)
    bootstrap_config = asyncio.run(cli_ctx.get_bootstrap_config())

    async def _clear_redis_history() -> None:
        try:
            async with connect_database(bootstrap_config.db) as db:
                async with db.begin_readonly() as conn:
                    query = (
                        sa.select(kernels.c.id)
                        .select_from(kernels)
                        .where(
                            (kernels.c.terminated_at < expiration_date),
                        )
                    )
                    result = await conn.execute(query)
                    target_kernels = [str(x.id) for x in result.all()]

            delete_count = 0
            async with redis_ctx(cli_ctx) as redis_conn_set:
                if len(target_kernels) > 0:
                    # Apply chunking to avoid excessive length of command params
                    # and indefinite blocking of the Redis server.
                    for kernel_ids in chunked(target_kernels, 32):
                        deleted = await redis_conn_set.stat.delete(kernel_ids)
                        delete_count += deleted
                    log.info(
                        "Cleaned up {:,} redis statistics records older than {:}.",
                        delete_count,
                        expiration_date,
                    )

                # Sync and compact the persistent database of Redis
                redis_config = await redis_conn_set.stat.config_get(["appendonly"])
                if redis_config.get(b"appendonly") == b"yes":
                    await redis_conn_set.stat.execute_command(["BGREWRITEAOF"])
                    log.info("Issued BGREWRITEAOF to the Redis database.")
                else:
                    await redis_conn_set.stat.execute_command(["BGSAVE", "SCHEDULE"])
                    log.info("Issued BGSAVE to the Redis database.")
        except Exception:
            log.exception("Unexpected error while cleaning up redis history")

    async def _clear_terminated_sessions() -> None:
        async with connect_database(bootstrap_config.db, isolation_level="AUTOCOMMIT") as db:
            async with db.begin() as conn:
                log.info("Deleting old records...")
                result = (
                    await conn.scalars(
                        sa.select(SessionRow.id).where(SessionRow.terminated_at < expiration_date)
                    )
                ).all()
                session_ids = cast(list[uuid.UUID], result)
                if session_ids:
                    await conn.execute(
                        sa.delete(kernels).where(kernels.c.session_id.in_(session_ids))
                    )
                    await conn.execute(sa.delete(SessionRow).where(SessionRow.id.in_(session_ids)))

                curs = await conn.execute(sa.select(sa.func.count()).select_from(SessionRow))
                if ret := curs.fetchone():
                    table_size = ret[0]
                    log.info(
                        "The number of rows of the `sessions` tables after cleanup: {}",
                        table_size,
                    )
        log.info(
            "Cleaned up {:,} database records older than {}.",
            len(session_ids),
            expiration_date,
        )

    async def _clear_old_error_logs() -> None:
        async with connect_database(bootstrap_config.db, isolation_level="AUTOCOMMIT") as db:
            async with db.begin() as conn:
                log.info("Deleting old error logs...")
                result = await conn.execute(
                    sa.delete(error_logs).where(error_logs.c.created_at < expiration_date),
                )
                deleted_count = result.rowcount

        log.info(
            "Cleaned up {:,} error log records older than {}.",
            deleted_count,
            expiration_date,
        )

    asyncio.run(_clear_redis_history())
    asyncio.run(_clear_terminated_sessions())
    asyncio.run(_clear_old_error_logs())
    asyncio.run(vacuum_db(bootstrap_config, vacuum_full))


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.dbschema:cli")
def schema() -> None:
    """Command set for managing the database schema."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.etcd:cli")
def etcd() -> None:
    """Command set for putting/getting data to/from etcd."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.fixture:cli")
def fixture() -> None:
    """Command set for managing fixtures."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.api:cli")
def api() -> None:
    """Command set for API schema inspection and manipulation."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.gql:cli")
def gql() -> None:
    """Command set for GraphQL schema."""
    # Deprecated in favor of "api" but kept for backward compatibility


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.image:cli")
def image() -> None:
    """Command set for managing images."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.redis:cli")
def redis() -> None:
    """Command set for Redis related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.agent:cli")
def agent() -> None:
    """Command set for agent related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.scheduler:cli")
def scheduler() -> None:
    """Command set for scheduler related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.config:cli")
def config() -> None:
    """Command set for configuration management."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.dependencies:cli")
def dependencies() -> None:
    """Command set for dependency verification and validation."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.health:cli")
def health() -> None:
    """Command set for health checking."""


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from ai.backend.manager.api.utils import Undefined


def drop_undefined(d: dict[Any, Any]) -> dict[Any, Any]:
    newd: dict[Any, Any] = {}
    for k, v in d.items():
        if isinstance(v, (Mapping, dict)):
            newval = drop_undefined(dict(v))
            if len(newval.keys()) > 0:  # exclude empty dict always
                newd[k] = newval
        elif not isinstance(v, Undefined):
            newd[k] = v
    return newd



from __future__ import annotations

import logging
import sys
import traceback
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, override

from ai.backend.common.events.event_types.agent.anycast import AgentErrorEvent
from ai.backend.common.events.types import AbstractEvent
from ai.backend.common.plugin.event import AbstractEventDispatcherPlugin
from ai.backend.common.plugin.monitor import AbstractErrorReporterPlugin
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter, LogLevel
from ai.backend.manager.data.error_log.types import ErrorLogSeverity
from ai.backend.manager.repositories.base import Creator
from ai.backend.manager.repositories.error_log.creators import ErrorLogCreatorSpec

if TYPE_CHECKING:
    from ai.backend.manager.api.context import RootContext
    from ai.backend.manager.repositories.error_log import ErrorLogRepository

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ErrorMonitor(AbstractErrorReporterPlugin):
    _error_log_repository: ErrorLogRepository

    async def init(self, context: Any | None = None) -> None:
        if context is None:
            log.warning(
                "manager.plugin.error_monitor is initialized without the root context. "
                "The plugin is disabled.",
            )
            self.enabled = False
            return
        self.enabled = True
        root_ctx: RootContext = context["_root.context"]
        self._error_log_repository = root_ctx.repositories.error_log.repository

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        pass

    async def capture_message(self, message: str) -> None:
        pass

    async def capture_exception(
        self,
        exc_instance: Exception | None = None,
        context: Mapping[str, Any] | None = None,
    ) -> None:
        if not self.enabled:
            return
        if exc_instance:
            tb = exc_instance.__traceback__
        else:
            _, sys_exc_instance, tb = sys.exc_info()
            if isinstance(sys_exc_instance, BaseException) and not isinstance(
                sys_exc_instance, Exception
            ):
                # bypass BaseException as they are used for controlling the process/coroutine lifecycles
                # instead of indicating actual errors
                return
            exc_instance = sys_exc_instance
        exc_type: Any = type(exc_instance)

        if context is None or "severity" not in context:
            severity = LogLevel.ERROR
        else:
            severity = context["severity"]
        if context is None or "user" not in context:
            user = None
        else:
            user = context["user"]
        message = "".join(traceback.format_exception_only(exc_type, exc_instance)).strip()

        error_log_severity = ErrorLogSeverity(severity.value.lower())
        creator = Creator(
            spec=ErrorLogCreatorSpec(
                severity=error_log_severity,
                source="manager",
                user=user,
                message=message,
                context_lang="python",
                context_env=dict(context) if context else {},
                traceback="".join(traceback.format_tb(tb)).strip(),
            )
        )
        await self._error_log_repository.create(creator)
        log.debug(
            'collected an error log [{}] "{}" from manager',
            severity.name,
            message,
        )


class ErrorEventDispatcher(AbstractEventDispatcherPlugin):
    _error_log_repository: ErrorLogRepository

    async def init(self, context: Any | None = None) -> None:
        if context is None:
            log.warning(
                "manager.plugin.error_event_dispatcher is initialized without the root context. "
                "The plugin is disabled.",
            )
            self.enabled = False
            return
        self.enabled = True
        root_ctx: RootContext = context
        self._error_log_repository = root_ctx.repositories.error_log.repository

    @override
    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        pass

    @override
    async def handle_event(
        self,
        source: AgentId,
        event: AbstractEvent,
    ) -> None:
        if not isinstance(event, AgentErrorEvent):
            return
        if not self.enabled:
            return
        error_log_severity = ErrorLogSeverity(event.severity.value.lower())
        creator = Creator(
            spec=ErrorLogCreatorSpec(
                severity=error_log_severity,
                source=source,
                user=event.user,
                message=event.message,
                context_lang="python",
                context_env=dict(event.context_env) if event.context_env else {},
                traceback=event.traceback,
            )
        )
        await self._error_log_repository.create(creator)
        log.debug(
            'collected an error log [{}] "{}" from agent:{}',
            event.severity.name,
            event.message,
            source,
        )



from .row import (
    MAXIMUM_DOTFILE_SIZE,
    AssocGroupUserRow,
    GroupDotfile,
    GroupRow,
    ProjectType,
    association_groups_users,
    groups,
    query_group_domain,
    query_group_dotfiles,
    resolve_group_name_or_id,
    verify_dotfile_name,
)
from .row import (
    get_permission_ctx as get_permission_ctx,
)
from .row import (
    resolve_groups as resolve_groups,
)

# __all__ controls what gets exported with "from .group import *"
# Keep it matching original to avoid conflicts with domain's get_permission_ctx in models/__init__.py
__all__ = (
    "MAXIMUM_DOTFILE_SIZE",
    "AssocGroupUserRow",
    "GroupDotfile",
    "GroupRow",
    "ProjectType",
    "association_groups_users",
    "groups",
    "query_group_domain",
    "query_group_dotfiles",
    "resolve_group_name_or_id",
    "verify_dotfile_name",
)



"""Type definitions for repository layer."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeVar

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError

if TYPE_CHECKING:
    from sqlalchemy.engine import Row

# QueryCondition now returns a ColumnElement (whereclause) instead of modifying stmt
type QueryCondition = Callable[[], sa.sql.expression.ColumnElement[bool]]

T = TypeVar("T")


@dataclass(frozen=True)
class ExistenceCheck[T]:
    """Defines an existence check for scope validation.

    Used to validate that required entities exist before executing a query.
    Multiple checks are combined into a single query for efficiency.
    """

    column: sa.orm.attributes.InstrumentedAttribute[T]
    """The column to check (e.g., ScalingGroupRow.name)."""

    value: T
    """The value to check for existence."""

    error: BackendAIError
    """The error to raise if the entity doesn't exist."""


class SearchScope(ABC):
    """Abstract base class for search scope.

    Scope defines required parameters for entity-based search queries.
    It converts to a QueryCondition that can be added to BatchQuerier conditions.
    Optionally defines existence checks for validation.
    """

    @abstractmethod
    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition."""
        raise NotImplementedError

    @property
    @abstractmethod
    def existence_checks(self) -> Sequence[ExistenceCheck[Any]]:
        """Return existence checks for scope validation.

        All checks are validated in a single query before the main query executes.
        """
        raise NotImplementedError


type QueryOrder = sa.sql.expression.UnaryExpression[Any] | sa.sql.expression.ColumnElement[Any]

# Factory function that creates a cursor condition from a decoded cursor value (str or UUID)
type CursorConditionFactory = Callable[[str], QueryCondition]

TRow = TypeVar("TRow", bound="Row[Any]")



"""User GraphQL mutation resolvers."""

from __future__ import annotations

from typing import cast
from uuid import UUID

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.user.types import (
    BulkCreateUserErrorGQL,
    BulkCreateUserInputGQL,
    BulkCreateUsersPayloadGQL,
    CreateUserInputGQL,
    CreateUserPayloadGQL,
    DeleteUserPayloadGQL,
    DeleteUsersInputGQL,
    DeleteUsersPayloadGQL,
    PurgeUserInputGQL,
    PurgeUserPayloadGQL,
    PurgeUsersInputGQL,
    PurgeUsersPayloadGQL,
    UpdateUserInputGQL,
    UpdateUserPayloadGQL,
    UserV2GQL,
)
from ai.backend.manager.data.user.types import UserStatus
from ai.backend.manager.models.hasher.types import PasswordInfo
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.user.creators import UserCreatorSpec
from ai.backend.manager.services.user.actions.create_user import (
    BulkCreateUserAction,
    UserCreateSpec,
)

# Create Mutations


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Create a new user (admin only). "
        "Requires superadmin privileges. "
        "Automatically creates a default keypair for the user."
    )
)  # type: ignore[misc]
async def admin_create_user(
    info: Info[StrawberryGQLContext],
    input: CreateUserInputGQL,
) -> CreateUserPayloadGQL:
    """Create a new user.

    Args:
        info: Strawberry GraphQL context.
        input: User creation input.

    Returns:
        CreateUserPayloadGQL with the created user.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_create_user is not yet implemented")


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Create multiple users in bulk (admin only). "
        "Requires superadmin privileges. "
        "Each user has individual specifications."
    )
)  # type: ignore[misc]
async def admin_bulk_create_users(
    info: Info[StrawberryGQLContext],
    input: BulkCreateUserInputGQL,
) -> BulkCreateUsersPayloadGQL:
    """Create multiple users in bulk with individual specifications.

    Args:
        info: Strawberry GraphQL context.
        input: Bulk user creation input with individual specs.

    Returns:
        BulkCreateUsersPayloadGQL with created users.
    """
    ctx = info.context
    auth_config = ctx.config_provider.config.auth

    # Build list of UserCreateSpec from input
    items: list[UserCreateSpec] = []
    for user_input in input.users:
        password_info = PasswordInfo(
            password=user_input.password,
            algorithm=auth_config.password_hash_algorithm,
            rounds=auth_config.password_hash_rounds,
            salt_size=auth_config.password_hash_salt_size,
        )

        spec = UserCreatorSpec(
            email=user_input.email,
            username=user_input.username,
            password=password_info,
            need_password_change=user_input.need_password_change,
            domain_name=user_input.domain_name,
            full_name=user_input.full_name,
            description=user_input.description,
            status=UserStatus(user_input.status.value),
            role=user_input.role.value,
            allowed_client_ip=user_input.allowed_client_ip,
            totp_activated=user_input.totp_activated,
            resource_policy=user_input.resource_policy,
            sudo_session_enabled=user_input.sudo_session_enabled,
            container_uid=user_input.container_uid,
            container_main_gid=user_input.container_main_gid,
            container_gids=user_input.container_gids,
        )

        group_ids = [str(gid) for gid in user_input.group_ids] if user_input.group_ids else None
        items.append(UserCreateSpec(creator=Creator(spec=spec), group_ids=group_ids))

    action = BulkCreateUserAction(items=items)
    result = await ctx.processors.user.bulk_create_users.wait_for_complete(action)

    created_users = [UserV2GQL.from_data(user_data) for user_data in result.data.successes]
    failed = [
        BulkCreateUserErrorGQL(
            index=error.index,
            username=cast(UserCreatorSpec, error.spec).username,
            email=cast(UserCreatorSpec, error.spec).email,
            message=str(error.exception),
        )
        for error in result.data.failures
    ]

    return BulkCreateUsersPayloadGQL(
        created_users=created_users,
        failed=failed,
    )


# Update Mutations


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Update a user's information (admin only). "
        "Requires superadmin privileges. "
        "Only provided fields will be updated."
    )
)  # type: ignore[misc]
async def admin_update_user(
    info: Info[StrawberryGQLContext],
    user_id: UUID,
    input: UpdateUserInputGQL,
) -> UpdateUserPayloadGQL:
    """Update a user's information.

    Args:
        info: Strawberry GraphQL context.
        user_id: UUID of the user to update.
        input: User update input with fields to modify.

    Returns:
        UpdateUserPayloadGQL with the updated user.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_update_user is not yet implemented")


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Update the current user's information. "
        "Users can only update their own profile. "
        "Some fields may be restricted based on user role."
    )
)  # type: ignore[misc]
async def update_user(
    info: Info[StrawberryGQLContext],
    input: UpdateUserInputGQL,
) -> UpdateUserPayloadGQL:
    """Update the current user's own information.

    Args:
        info: Strawberry GraphQL context.
        input: User update input with fields to modify.

    Returns:
        UpdateUserPayloadGQL with the updated user.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("update_user is not yet implemented")


# Delete Mutations (Soft Delete)


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Soft-delete a user (admin only). "
        "Requires superadmin privileges. "
        "Sets the user status to DELETED but preserves data."
    )
)  # type: ignore[misc]
async def admin_delete_user(
    info: Info[StrawberryGQLContext],
    user_id: UUID,
) -> DeleteUserPayloadGQL:
    """Soft-delete a single user.

    Args:
        info: Strawberry GraphQL context.
        user_id: UUID of the user to delete.

    Returns:
        DeleteUserPayloadGQL indicating success.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_delete_user is not yet implemented")


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Soft-delete multiple users (admin only). "
        "Requires superadmin privileges. "
        "Sets user status to DELETED but preserves data."
    )
)  # type: ignore[misc]
async def admin_delete_users(
    info: Info[StrawberryGQLContext],
    input: DeleteUsersInputGQL,
) -> DeleteUsersPayloadGQL:
    """Soft-delete multiple users.

    Args:
        info: Strawberry GraphQL context.
        input: Input containing list of user UUIDs to delete.

    Returns:
        DeleteUsersPayloadGQL with count of deleted users.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_delete_users is not yet implemented")


# Purge Mutations (Hard Delete)


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Permanently delete a user and all associated data (admin only). "
        "Requires superadmin privileges. "
        "This action is IRREVERSIBLE. All user data, sessions, and resources will be deleted."
    )
)  # type: ignore[misc]
async def admin_purge_user(
    info: Info[StrawberryGQLContext],
    input: PurgeUserInputGQL,
) -> PurgeUserPayloadGQL:
    """Permanently delete a single user.

    Args:
        info: Strawberry GraphQL context.
        input: Purge input with email and options.

    Returns:
        PurgeUserPayloadGQL indicating success.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_purge_user is not yet implemented")


@strawberry.mutation(
    description=(
        "Added in 26.2.0. Permanently delete multiple users (admin only). "
        "Requires superadmin privileges. "
        "This action is IRREVERSIBLE. All user data will be deleted."
    )
)  # type: ignore[misc]
async def admin_purge_users(
    info: Info[StrawberryGQLContext],
    input: PurgeUsersInputGQL,
) -> PurgeUsersPayloadGQL:
    """Permanently delete multiple users.

    Args:
        info: Strawberry GraphQL context.
        input: Purge input with emails and options.

    Returns:
        PurgeUsersPayloadGQL with count of purged users.

    Raises:
        NotImplementedError: This mutation is not yet implemented.
    """
    raise NotImplementedError("admin_purge_users is not yet implemented")



from abc import abstractmethod
from collections.abc import Sequence

from ai.backend.manager.data.deployment.types import DeploymentInfo, DeploymentStatusTransitions
from ai.backend.manager.data.model_serving.types import EndpointLifecycle
from ai.backend.manager.defs import LockID
from ai.backend.manager.sokovan.deployment.types import DeploymentExecutionResult


class DeploymentHandler:
    """Base class for deployment operation handlers using the generic interface."""

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        raise NotImplementedError("Subclasses must implement name()")

    @property
    @abstractmethod
    def lock_id(self) -> LockID | None:
        """Get the lock ID for this handler.

        Returns:
            LockID to acquire before execution, or None if no lock needed
        """
        raise NotImplementedError("Subclasses must implement lock_id")

    @classmethod
    @abstractmethod
    def target_statuses(cls) -> list[EndpointLifecycle]:
        """Get the target deployment statuses for this handler.

        Returns:
            List of deployment statuses that this handler targets
        """
        raise NotImplementedError("Subclasses must implement target_statuses()")

    @classmethod
    @abstractmethod
    def next_status(cls) -> EndpointLifecycle | None:
        """Get the next deployment status after this handler's operation.

        Returns:
            The next deployment status
        """
        raise NotImplementedError("Subclasses must implement next_status()")

    @classmethod
    @abstractmethod
    def failure_status(cls) -> EndpointLifecycle | None:
        """Get the failure deployment status if applicable.

        Returns:
            The failure deployment status, or None if not applicable
        """
        raise NotImplementedError("Subclasses must implement failure_status()")

    @classmethod
    @abstractmethod
    def status_transitions(cls) -> DeploymentStatusTransitions:
        """Define state transitions for different handler outcomes (BEP-1030).

        Returns:
            DeploymentStatusTransitions defining what lifecycle to transition to for
            success and failure outcomes.

        Note:
            - None value: Don't change the deployment lifecycle
        """
        raise NotImplementedError("Subclasses must implement status_transitions()")

    @abstractmethod
    async def execute(self, deployments: Sequence[DeploymentInfo]) -> DeploymentExecutionResult:
        """Execute the scheduling operation.

        Returns:
            Result of the scheduling operation
        """
        raise NotImplementedError("Subclasses must implement execute()")

    @abstractmethod
    async def post_process(self, result: DeploymentExecutionResult) -> None:
        """Handle post-processing after the operation.

        Args:
            result: The result from execute()
        """
        raise NotImplementedError("Subclasses must implement post_process()")



"""Deployment fetcher functions."""

from __future__ import annotations

from functools import lru_cache
from uuid import UUID

from strawberry import Info
from strawberry.relay import PageInfo

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.deployment.types.deployment import (
    DeploymentFilter,
    DeploymentOrderBy,
    ModelDeployment,
    ModelDeploymentConnection,
    ModelDeploymentEdge,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.endpoint.row import EndpointRow
from ai.backend.manager.repositories.deployment.options import (
    DeploymentConditions,
    DeploymentOrders,
)
from ai.backend.manager.services.deployment.actions.search_deployments import (
    SearchDeploymentsAction,
)


@lru_cache(maxsize=1)
def get_deployment_pagination_spec() -> PaginationSpec:
    return PaginationSpec(
        forward_order=DeploymentOrders.created_at(ascending=False),
        backward_order=DeploymentOrders.created_at(ascending=True),
        forward_condition_factory=DeploymentConditions.by_cursor_forward,
        backward_condition_factory=DeploymentConditions.by_cursor_backward,
        tiebreaker_order=EndpointRow.id.asc(),
    )


async def fetch_deployments(
    info: Info[StrawberryGQLContext],
    filter: DeploymentFilter | None = None,
    order_by: list[DeploymentOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ModelDeploymentConnection:
    """Fetch deployments with optional filtering, ordering, and pagination."""
    processor = info.context.processors.deployment

    # Build querier using gql_adapter
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_deployment_pagination_spec(),
        filter=filter,
        order_by=order_by,
    )

    action_result = await processor.search_deployments.wait_for_complete(
        SearchDeploymentsAction(querier=querier)
    )

    nodes = [ModelDeployment.from_dataclass(data) for data in action_result.data]
    edges = [ModelDeploymentEdge(node=node, cursor=encode_cursor(str(node.id))) for node in nodes]

    return ModelDeploymentConnection(
        edges=edges,
        page_info=PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )


async def fetch_deployment(
    info: Info[StrawberryGQLContext],
    deployment_id: UUID,
) -> ModelDeployment | None:
    """Fetch a specific deployment by ID."""
    deployment_data = await info.context.data_loaders.deployment_loader.load(deployment_id)
    if deployment_data is None:
        return None
    return ModelDeployment.from_dataclass(deployment_data)



import logging
from io import BytesIO

import sqlalchemy as sa

from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.events.event_types.kernel.anycast import (
    DoSyncKernelLogsEvent,
    KernelCancelledAnycastEvent,
    KernelCreatingAnycastEvent,
    KernelPreparingAnycastEvent,
    KernelPullingAnycastEvent,
    KernelStartedAnycastEvent,
    KernelTerminatedAnycastEvent,
    KernelTerminatingAnycastEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.models.kernel import kernels
from ai.backend.manager.models.utils import (
    ExtendedAsyncSAEngine,
    execute_with_retry,
)
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class KernelEventHandler:
    _valkey_container_log: ValkeyContainerLogClient
    _valkey_stat: ValkeyStatClient
    _valkey_stream: ValkeyStreamClient
    _registry: AgentRegistry
    _db: ExtendedAsyncSAEngine
    _schedule_coordinator: ScheduleCoordinator

    def __init__(
        self,
        valkey_container_log: ValkeyContainerLogClient,
        valkey_stat: ValkeyStatClient,
        valkey_stream: ValkeyStreamClient,
        registry: AgentRegistry,
        db: ExtendedAsyncSAEngine,
        schedule_coordinator: ScheduleCoordinator,
    ) -> None:
        self._valkey_container_log = valkey_container_log
        self._valkey_stat = valkey_stat
        self._valkey_stream = valkey_stream
        self._registry = registry
        self._db = db
        self._schedule_coordinator = schedule_coordinator

    async def handle_kernel_log(
        self,
        _context: None,
        _source: AgentId,
        event: DoSyncKernelLogsEvent,
    ) -> None:
        # The log data is at most 10 MiB.
        log_buffer = BytesIO()
        try:
            list_size = await self._valkey_container_log.container_log_len(
                container_id=event.container_id
            )
            for _ in range(list_size):
                # Read chunk-by-chunk to allow interleaving with other Redis operations.
                chunks = await self._valkey_container_log.pop_container_logs(
                    container_id=event.container_id
                )
                if chunks is None:  # maybe missing
                    log_buffer.write(b"(container log unavailable)\n")
                    break
                for chunk in chunks:
                    log_buffer.write(chunk.get_content())
            try:
                log_data = log_buffer.getvalue()

                async def _update_log() -> None:
                    async with self._db.begin() as conn:
                        update_query = (
                            sa.update(kernels)
                            .values(container_log=log_data)
                            .where(kernels.c.id == event.kernel_id)
                        )
                        await conn.execute(update_query)

                await execute_with_retry(_update_log)
            finally:
                # Clear the log data from Redis when done.
                await self._valkey_container_log.clear_container_logs(
                    container_id=event.container_id
                )
        except Exception:
            # skip all exception in handle_kernel_log
            pass
        finally:
            log_buffer.close()

    async def handle_kernel_preparing(
        self,
        _context: None,
        _source: AgentId,
        event: KernelPreparingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_preparing: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        await self._schedule_coordinator.handle_kernel_preparing(event)

    async def handle_kernel_pulling(
        self,
        _context: None,
        _source: AgentId,
        event: KernelPullingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_pulling: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        await self._schedule_coordinator.handle_kernel_pulling(event)

    async def handle_kernel_creating(
        self,
        _context: None,
        _source: AgentId,
        event: KernelCreatingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_creating: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        await self._schedule_coordinator.handle_kernel_creating(event)

    async def handle_kernel_started(
        self,
        _context: None,
        _source: AgentId,
        event: KernelStartedAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_started: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        await self._schedule_coordinator.handle_kernel_running(event)

    async def handle_kernel_cancelled(
        self,
        _context: None,
        _source: AgentId,
        event: KernelCancelledAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_cancelled: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        await self._schedule_coordinator.handle_kernel_cancelled(event)

    async def handle_kernel_terminating(
        self,
        context: None,
        source: AgentId,
        event: KernelTerminatingAnycastEvent,
    ) -> None:
        # `destroy_kernel()` has already changed the kernel status to "TERMINATING".
        # No additional handling needed for terminating state
        pass

    async def handle_kernel_terminated(
        self,
        _context: None,
        _source: AgentId,
        event: KernelTerminatedAnycastEvent,
    ) -> None:
        await self._schedule_coordinator.handle_kernel_terminated(event)



from __future__ import annotations

import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import Any, override

import aiohttp
import sqlalchemy as sa
import yarl

from ai.backend.common.docker import arch_name_aliases, get_docker_connector
from ai.backend.common.json import pretty_json_str
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.image.types import ImageStatus
from ai.backend.manager.models.image import ImageRow

from .base import (
    BaseContainerRegistry,
    concurrency_sema,
    progress_reporter,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class LocalRegistry(BaseContainerRegistry):
    @actxmgr
    async def prepare_client_session(self) -> AsyncIterator[tuple[yarl.URL, aiohttp.ClientSession]]:
        connector = get_docker_connector()
        async with aiohttp.ClientSession(connector=connector.connector) as sess:
            yield connector.docker_host, sess

    @override
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        async with sess.get(self.registry_url / "images" / "json") as response:
            items = await response.json()
            if (reporter := progress_reporter.get()) is not None:
                reporter.total_progress = len(items)
            for item in items:
                if item["RepoTags"] is not None:
                    for image_ref_str in item["RepoTags"]:
                        if image_ref_str == "<none>:<none>":
                            # cache images
                            continue
                        yield image_ref_str  # this includes the tag part

    @override
    async def _scan_image(
        self,
        sess: aiohttp.ClientSession,
        image: str,
    ) -> None:
        repo, _, tag = image.rpartition(":")
        await self._scan_tag_local(sess, {}, repo, tag)

    async def _scan_tag_local(
        self,
        sess: aiohttp.ClientSession,
        _rqst_args: dict[str, str],
        image: str,
        tag: str,
    ) -> None:
        async def _read_image_info(
            _tag: str,
        ) -> tuple[dict[str, dict[str, Any]], str | None]:
            async with sess.get(
                self.registry_url / "images" / f"{image}:{tag}" / "json"
            ) as response:
                data = await response.json()
            architecture = arch_name_aliases.get(data["Architecture"], data["Architecture"])
            summary = {
                "Id": data["Id"],
                "RepoDigests": data.get("RepoDigests", []),
                "Config.Image": data["Config"]["Image"],
                "ContainerConfig.Image": data.get("ContainerConfig", {}).get("Image", None),
                "Architecture": architecture,
            }
            log.debug("scanned image info: {}:{}\n{}", image, tag, pretty_json_str(summary))
            config_digest = data["Id"]
            async with self.db.begin_readonly_session() as db_session:
                already_exists = await db_session.scalar(
                    sa.select(sa.func.count(ImageRow.id)).where(
                        sa.and_(
                            ImageRow.config_digest == config_digest,
                            ImageRow.is_local == sa.false(),
                            ImageRow.status == ImageStatus.ALIVE,
                        )
                    ),
                )
            if already_exists is not None and already_exists > 0:
                return {}, "already synchronized from a remote registry"
            labels = data["Config"]["Labels"]
            if labels is None:
                log.debug(
                    "The image {}:{}/{} has no metadata labels -> treating as vanilla image",
                    image,
                    tag,
                    architecture,
                )
                labels = {}
            return {
                architecture: {
                    "size": data["Size"],
                    "labels": labels,
                    "digest": config_digest,
                },
            }, None

        async with concurrency_sema.get():
            manifests, skip_reason = await _read_image_info(tag)
            await self._read_manifest(image, tag, manifests, skip_reason)



"""UpdaterSpec implementations for reservoir registry repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.reservoir_registry import ReservoirRegistryRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState


@dataclass
class ReservoirRegistryUpdaterSpec(UpdaterSpec[ReservoirRegistryRow]):
    """UpdaterSpec for reservoir registry updates."""

    endpoint: OptionalState[str] = field(default_factory=OptionalState.nop)
    access_key: OptionalState[str] = field(default_factory=OptionalState.nop)
    secret_key: OptionalState[str] = field(default_factory=OptionalState.nop)
    api_version: OptionalState[str] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[ReservoirRegistryRow]:
        return ReservoirRegistryRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.endpoint.update_dict(to_update, "endpoint")
        self.access_key.update_dict(to_update, "access_key")
        self.secret_key.update_dict(to_update, "secret_key")
        self.api_version.update_dict(to_update, "api_version")
        return to_update



from dataclasses import dataclass
from typing import override

from ai.backend.common.bgtask.reporter import ProgressReporter
from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.services.container_registry.actions.base import ContainerRegistryAction


@dataclass
class RescanImagesAction(ContainerRegistryAction):
    registry: str
    project: str | None
    progress_reporter: ProgressReporter | None

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.CONTAINER_REGISTRY_IMAGE

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class RescanImagesActionResult(BaseActionResult):
    images: list[ImageData]
    registry: ContainerRegistryData
    errors: list[str]

    @override
    def entity_id(self) -> str | None:
        return str(self.registry.id)



"""Creator specs for Fair Share repository INSERT operations."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from decimal import Decimal
from typing import override

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.models.fair_share import (
    DomainFairShareRow,
    ProjectFairShareRow,
    UserFairShareRow,
)
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class DomainFairShareCreatorSpec(CreatorSpec[DomainFairShareRow]):
    """Creator spec for DomainFairShareRow.

    Only identity fields are required. Spec fields use Row defaults if not provided.
    """

    # Identity (required)
    resource_group: str
    domain_name: str

    # Spec (optional - uses Row defaults if None)
    weight: Decimal | None = None
    half_life_days: int | None = None
    lookback_days: int | None = None
    decay_unit_days: int | None = None
    resource_weights: ResourceSlot | None = None

    @override
    def build_row(self) -> DomainFairShareRow:
        row = DomainFairShareRow(
            resource_group=self.resource_group,
            domain_name=self.domain_name,
        )
        if self.weight is not None:
            row.weight = self.weight
        if self.half_life_days is not None:
            row.half_life_days = self.half_life_days
        if self.lookback_days is not None:
            row.lookback_days = self.lookback_days
        if self.decay_unit_days is not None:
            row.decay_unit_days = self.decay_unit_days
        if self.resource_weights is not None:
            row.resource_weights = self.resource_weights
        return row


@dataclass
class ProjectFairShareCreatorSpec(CreatorSpec[ProjectFairShareRow]):
    """Creator spec for ProjectFairShareRow.

    Only identity fields are required. Spec fields use Row defaults if not provided.
    """

    # Identity (required)
    resource_group: str
    project_id: uuid.UUID
    domain_name: str

    # Spec (optional - uses Row defaults if None)
    weight: Decimal | None = None
    half_life_days: int | None = None
    lookback_days: int | None = None
    decay_unit_days: int | None = None
    resource_weights: ResourceSlot | None = None

    @override
    def build_row(self) -> ProjectFairShareRow:
        row = ProjectFairShareRow(
            resource_group=self.resource_group,
            project_id=self.project_id,
            domain_name=self.domain_name,
        )
        if self.weight is not None:
            row.weight = self.weight
        if self.half_life_days is not None:
            row.half_life_days = self.half_life_days
        if self.lookback_days is not None:
            row.lookback_days = self.lookback_days
        if self.decay_unit_days is not None:
            row.decay_unit_days = self.decay_unit_days
        if self.resource_weights is not None:
            row.resource_weights = self.resource_weights
        return row


@dataclass
class UserFairShareCreatorSpec(CreatorSpec[UserFairShareRow]):
    """Creator spec for UserFairShareRow.

    Only identity fields are required. Spec fields use Row defaults if not provided.
    """

    # Identity (required)
    resource_group: str
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str

    # Spec (optional - uses Row defaults if None)
    weight: Decimal | None = None
    half_life_days: int | None = None
    lookback_days: int | None = None
    decay_unit_days: int | None = None
    resource_weights: ResourceSlot | None = None

    @override
    def build_row(self) -> UserFairShareRow:
        row = UserFairShareRow(
            resource_group=self.resource_group,
            user_uuid=self.user_uuid,
            project_id=self.project_id,
            domain_name=self.domain_name,
        )
        if self.weight is not None:
            row.weight = self.weight
        if self.half_life_days is not None:
            row.half_life_days = self.half_life_days
        if self.lookback_days is not None:
            row.lookback_days = self.lookback_days
        if self.decay_unit_days is not None:
            row.decay_unit_days = self.decay_unit_days
        if self.resource_weights is not None:
            row.resource_weights = self.resource_weights
        return row



"""
Exceptions related to dependency management.
"""

from __future__ import annotations

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class InvalidManagerKeypairError(BackendAIError):
    """Raised when manager secret key is missing from the keypair file."""

    error_type = "https://api.backend.ai/probs/dependency/invalid-manager-keypair"
    error_title = "Invalid manager keypair configuration."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.SETUP,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



"""Mount data preparation rule."""

from typing import Any

from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)

from .base import SessionPreparerRule


class MountPreparationRule(SessionPreparerRule):
    """Prepares mount data for the session."""

    def name(self) -> str:
        return "mount_preparation"

    def prepare(
        self,
        spec: SessionCreationSpec,
        _context: SessionCreationContext,
        preparation_data: dict[str, Any],
    ) -> None:
        """Prepare mount data."""
        mount_map = spec.creation_spec.get("mount_map") or {}
        mount_id_map = spec.creation_spec.get("mount_id_map") or {}
        combined_mount_map = {**mount_map, **mount_id_map}

        # Validation is now handled by MountNameValidationRule
        preparation_data["mount_map_prepared"] = True
        preparation_data["combined_mount_map"] = combined_mount_map



from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import datetime

from .role import RoleData


@dataclass
class UserRoleUpdateInput:
    pass


@dataclass
class UserRoleDeleteInput:
    id: uuid.UUID


@dataclass
class UserRoleData:
    id: uuid.UUID
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: uuid.UUID | None
    granted_at: datetime
    expires_at: datetime | None
    deleted_at: datetime | None


@dataclass
class UserRoleDataWithRole:
    id: uuid.UUID
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: uuid.UUID | None
    granted_at: datetime
    expires_at: datetime | None
    deleted_at: datetime | None

    mapped_role_data: RoleData



"""Streaming export types and utilities for server-side CSV export.

This module provides the core types and functions for streaming export functionality.
Based on BEP-1025: Server-Side CSV Export API.
"""

from __future__ import annotations

from collections.abc import AsyncIterator, Callable, Sequence
from dataclasses import dataclass
from enum import StrEnum
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa

from .types import QueryCondition, QueryOrder

if TYPE_CHECKING:
    from sqlalchemy.orm.attributes import InstrumentedAttribute

    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


class ExportLimitExceeded(Exception):
    """Raised when export row limit is exceeded."""

    pass


class ExportFieldType(StrEnum):
    """Export field type for client display and validation."""

    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    UUID = "uuid"
    JSON = "json"
    ENUM = "enum"


# Formatter type definition - converts any value to string for CSV
type ExportFormatter = Callable[[Any], str]


@dataclass(frozen=True)
class JoinDef:
    """JOIN definition for export queries.

    Defines a table and condition for LEFT JOIN operations.
    Used by ExportFieldDef to declare required joins for a field.

    Attributes:
        table: Table to join (e.g., ProjectResourcePolicyRow.__table__)
        condition: JOIN condition as SQLAlchemy expression
    """

    table: sa.FromClause
    condition: sa.ColumnElement[bool]


@dataclass(frozen=True)
class ExportFieldDef:
    """Export field definition.

    Informs clients which fields can be selected,
    and used for actual data transformation.

    Attributes:
        key: Field identifier (used in API requests)
        name: Name displayed in CSV header
        description: Field description (displayed in client UI)
        field_type: Field type for client display
        column: ORM column reference for SELECT query building
        formatter: Custom formatter (None uses default conversion)
        joins: Set of JoinDef required to access this field (None = no joins needed)
    """

    key: str
    name: str
    description: str
    field_type: ExportFieldType
    column: InstrumentedAttribute[Any]
    formatter: ExportFormatter | None = None
    joins: tuple[JoinDef, ...] | frozenset[JoinDef] | None = None


@dataclass(frozen=True)
class ExportQueryParams:
    """Input parameters for export query (without fields).

    Used by API layer to pass query parameters to the service.
    Service constructs the full StreamingExportQuery by adding fields.

    Attributes:
        conditions: WHERE clause conditions
        orders: ORDER BY clause sort conditions
        max_rows: Maximum export row count (passed from config.export.max_rows)
        statement_timeout_sec: DB statement timeout (passed from config.export.statement_timeout_sec)
    """

    conditions: list[QueryCondition]
    orders: list[QueryOrder]
    max_rows: int
    statement_timeout_sec: int


@dataclass(frozen=True)
class StreamingExportQuery:
    """Query definition for streaming export.

    Similar to BatchQuerier but streams all data without pagination.
    Limits export size with max_rows.

    Attributes:
        select_from: FROM clause (single table or joined tables)
        fields: Selected field definitions for SELECT query building
        conditions: WHERE clause conditions
        orders: ORDER BY clause sort conditions
        max_rows: Maximum export row count (passed from config.export.max_rows)
        statement_timeout_sec: DB statement timeout (passed from config.export.statement_timeout_sec)
    """

    select_from: sa.FromClause
    fields: list[ExportFieldDef]
    conditions: list[QueryCondition]
    orders: list[QueryOrder]
    max_rows: int
    statement_timeout_sec: int

    @classmethod
    def from_params(
        cls,
        params: ExportQueryParams,
        select_from: sa.FromClause,
        fields: list[ExportFieldDef],
    ) -> StreamingExportQuery:
        """Create StreamingExportQuery from params, select_from, and fields."""
        return cls(
            select_from=select_from,
            fields=fields,
            conditions=params.conditions,
            orders=params.orders,
            max_rows=params.max_rows,
            statement_timeout_sec=params.statement_timeout_sec,
        )


@dataclass
class ReportDef:
    """Export report definition.

    Defines available fields for export.

    Attributes:
        report_key: Report identifier (e.g., "audit-logs")
        name: Report name for display
        description: Report description
        select_from: FROM clause (single table or joined tables)
        fields: Available field list
    """

    report_key: str
    name: str
    description: str
    select_from: sa.FromClause
    fields: list[ExportFieldDef]

    def get_field(self, key: str) -> ExportFieldDef | None:
        """Find field definition by key."""
        for f in self.fields:
            if f.key == key:
                return f
        return None

    def get_field_keys(self) -> set[str]:
        """Return all field keys."""
        return {f.key for f in self.fields}

    def validate_field_keys(self, requested_keys: list[str]) -> list[str]:
        """Return invalid field keys."""
        valid_keys = self.get_field_keys()
        return [k for k in requested_keys if k not in valid_keys]

    def get_fields_by_keys(self, keys: list[str]) -> list[ExportFieldDef]:
        """Return field definitions for the given keys, preserving order."""
        key_to_field = {f.key: f for f in self.fields}
        return [key_to_field[k] for k in keys if k in key_to_field]


DEFAULT_PARTITION_SIZE = 1000


async def execute_streaming_export(
    db: ExtendedAsyncSAEngine,
    query: StreamingExportQuery,
    partition_size: int = DEFAULT_PARTITION_SIZE,
) -> AsyncIterator[Sequence[Sequence[Any]]]:
    """Execute streaming export based on query definition.

    Uses db_sess.stream() for server-side cursor data streaming.
    Builds SELECT query from field column references in query.fields.
    Yields rows in partitions for better performance.

    Args:
        db: Database engine
        query: Export query containing fields, conditions, orders, max_rows, statement_timeout_sec
        partition_size: Number of rows per partition (default: 1000)

    Yields:
        Partitions of row values (each row in query.fields order)
    """
    # Build SELECT from field columns with explicit FROM clause
    columns = [f.column for f in query.fields]
    base_query = sa.select(*columns).select_from(query.select_from)

    # Apply conditions
    for condition in query.conditions:
        base_query = base_query.where(condition())

    # Apply ordering
    for order in query.orders:
        base_query = base_query.order_by(order)

    # Apply max_rows LIMIT
    base_query = base_query.limit(query.max_rows)

    async with db.begin_readonly_session_read_committed() as db_sess:
        # Set statement timeout
        await db_sess.execute(sa.text(f"SET statement_timeout = '{query.statement_timeout_sec}s'"))

        # Server-side cursor streaming with partitions
        result = await db_sess.stream(base_query)

        async for partition in result.partitions(partition_size):
            yield [_transform_row(row, query.fields) for row in partition]


def _transform_row(
    row: sa.Row[Any],
    fields: list[ExportFieldDef],
) -> list[Any]:
    """Transform row to list of values in field order."""
    result: list[Any] = []
    for i, field_def in enumerate(fields):
        value = row[i]
        # Apply custom formatter if present
        if field_def.formatter is not None:
            value = field_def.formatter(value)
        result.append(value)
    return result


@dataclass(frozen=True)
class ExportDataStream:
    """Export data stream with field information.

    Bundles field names and data iterator for streaming export.
    Yields partitions of row values for efficient async processing.
    """

    field_names: list[str]
    reader: AsyncIterator[Sequence[Sequence[Any]]]  # partitions of row values



"""Route resolver functions."""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.manager.api.gql.base import resolve_global_id
from ai.backend.manager.api.gql.deployment.fetcher.route import fetch_routes
from ai.backend.manager.api.gql.deployment.types.route import (
    Route,
    RouteConnection,
    RouteFilter,
    RouteOrderBy,
    UpdateRouteTrafficStatusInputGQL,
    UpdateRouteTrafficStatusPayloadGQL,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.data.deployment.types import RouteTrafficStatus as RouteTrafficStatusEnum
from ai.backend.manager.repositories.deployment.options import RouteConditions
from ai.backend.manager.services.deployment.actions.route import (
    UpdateRouteTrafficStatusAction,
)

# Query resolvers


@strawberry.field(  # type: ignore[misc]
    description="Added in 25.19.0. List routes for a deployment with optional filters."
)
async def routes(
    info: Info[StrawberryGQLContext],
    deployment_id: ID,
    filter: RouteFilter | None = None,
    order_by: list[RouteOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> RouteConnection | None:
    """List routes for a deployment with optional filters."""
    _, endpoint_id = resolve_global_id(deployment_id)
    return await fetch_routes(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
        base_conditions=[RouteConditions.by_endpoint_id(UUID(endpoint_id))],
    )


@strawberry.field(description="Added in 25.19.0. Get a specific route by ID.")  # type: ignore[misc]
async def route(id: ID, info: Info[StrawberryGQLContext]) -> Route | None:
    """Get a specific route by ID."""
    _, route_id = resolve_global_id(id)

    route_info = await info.context.data_loaders.route_loader.load(UUID(route_id))
    if route_info is None:
        return None
    return Route.from_dataclass(route_info)


# Mutation resolvers


@strawberry.mutation(description="Added in 25.19.0. Update the traffic status of a route.")  # type: ignore[misc]
async def update_route_traffic_status(
    input: UpdateRouteTrafficStatusInputGQL,
    info: Info[StrawberryGQLContext],
) -> UpdateRouteTrafficStatusPayloadGQL:
    """Update route traffic status (ACTIVE/INACTIVE)."""
    _, route_id = resolve_global_id(input.route_id)

    processor = info.context.processors.deployment
    result = await processor.update_route_traffic_status.wait_for_complete(
        UpdateRouteTrafficStatusAction(
            route_id=UUID(route_id),
            traffic_status=RouteTrafficStatusEnum(input.traffic_status.value),
        )
    )

    return UpdateRouteTrafficStatusPayloadGQL(
        route=Route.from_dataclass(result.route),
    )



from __future__ import annotations

from decimal import Decimal
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

from ai.backend.common.types import ResourceSlot

__all__ = ("FairShareScalingGroupSpec",)


class FairShareScalingGroupSpec(BaseModel):
    """Fair Share calculation configuration for a Resource Group.

    Used for Fair Share metric calculation regardless of the scheduler type.
    """

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    half_life_days: int = 7
    """Half-life for exponential decay in days."""

    lookback_days: int = 28
    """Total lookback period in days for usage aggregation."""

    decay_unit_days: int = 1
    """Granularity of decay buckets in days."""

    default_weight: Decimal = Decimal("1.0")
    """Default weight for entities without explicit weight in this scaling group."""

    resource_weights: ResourceSlot = Field(default_factory=ResourceSlot)
    """Weights for each resource type when calculating normalized usage.

    If a resource type is not specified, default weight (1.0) is used.
    Example: ResourceSlot({"cpu": 1.0, "mem": 0.001, "cuda.device": 10.0})
    """

    @field_serializer("resource_weights", mode="plain")
    def serialize_resource_weights(self, value: ResourceSlot) -> dict[str, Any]:
        """Serialize ResourceSlot to dict for JSON compatibility."""
        return {k: str(v) for k, v in value.items()}

    @field_validator("resource_weights", mode="before")
    @classmethod
    def validate_resource_weights(cls, value: Any) -> ResourceSlot:
        """Deserialize dict to ResourceSlot.

        Converts string values to Decimal to avoid BinarySize parsing issues.
        """
        if isinstance(value, ResourceSlot):
            return value
        if isinstance(value, dict):
            # Convert string values to Decimal to bypass BinarySize parsing
            converted = {k: Decimal(v) if isinstance(v, str) else v for k, v in value.items()}
            return ResourceSlot(converted)
        return ResourceSlot()



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import NotificationAction


@dataclass
class ValidateChannelAction(NotificationAction):
    """Action to validate a notification channel (webhook sending test)."""

    channel_id: UUID
    test_message: str

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str | None:
        return str(self.channel_id)


@dataclass
class ValidateChannelActionResult(BaseActionResult):
    """Result of validating a notification channel."""

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, cast, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.notification import NotificationRuleData
from ai.backend.manager.repositories.base import Creator
from ai.backend.manager.repositories.notification.creators import NotificationRuleCreatorSpec

from .base import NotificationAction

if TYPE_CHECKING:
    from ai.backend.manager.models.notification import NotificationRuleRow


@dataclass
class CreateRuleAction(NotificationAction):
    """Action to create a notification rule."""

    creator: Creator[NotificationRuleRow]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE

    @override
    def entity_id(self) -> str | None:
        spec = cast(NotificationRuleCreatorSpec, self.creator.spec)
        return spec.name


@dataclass
class CreateRuleActionResult(BaseActionResult):
    """Result of creating a notification rule."""

    rule_data: NotificationRuleData

    @override
    def entity_id(self) -> str | None:
        return str(self.rule_data.id)



"""add entity_fields table

Revision ID: d86417a6bf7b
Revises: d0a3c0716970
Create Date: 2026-01-05 19:25:49.581618

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "d86417a6bf7b"
down_revision = "d0a3c0716970"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "entity_fields",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("entity_type", sa.String(length=64), nullable=False),
        sa.Column("entity_id", sa.String(length=64), nullable=False),
        sa.Column("field_type", sa.String(length=64), nullable=False),
        sa.Column("field_id", sa.String(length=64), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_entity_fields")),
        sa.UniqueConstraint(
            "entity_type",
            "entity_id",
            "field_type",
            "field_id",
            name="uq_entity_fields_mapping",
        ),
    )
    op.create_index(
        "ix_entity_fields_entity_lookup",
        "entity_fields",
        ["entity_type", "entity_id"],
    )
    op.create_index(
        "ix_entity_fields_field_lookup",
        "entity_fields",
        ["field_type", "field_id"],
    )


def downgrade() -> None:
    op.drop_index("ix_entity_fields_field_lookup", table_name="entity_fields")
    op.drop_index("ix_entity_fields_entity_lookup", table_name="entity_fields")
    op.drop_table("entity_fields")



"""Resource Usage GraphQL module."""

from __future__ import annotations

from .resolver import (
    admin_domain_usage_buckets,
    admin_project_usage_buckets,
    admin_user_usage_buckets,
    domain_usage_buckets,
    project_usage_buckets,
    rg_domain_usage_buckets,
    rg_project_usage_buckets,
    rg_user_usage_buckets,
    user_usage_buckets,
)
from .types import (
    DomainUsageBucketConnection,
    DomainUsageBucketEdge,
    DomainUsageBucketFilter,
    DomainUsageBucketGQL,
    DomainUsageBucketOrderBy,
    ProjectUsageBucketConnection,
    ProjectUsageBucketEdge,
    ProjectUsageBucketFilter,
    ProjectUsageBucketGQL,
    ProjectUsageBucketOrderBy,
    UsageBucketMetadataGQL,
    UsageBucketOrderField,
    UserUsageBucketConnection,
    UserUsageBucketEdge,
    UserUsageBucketFilter,
    UserUsageBucketGQL,
    UserUsageBucketOrderBy,
)

__all__ = (
    # Admin Resolvers
    "admin_domain_usage_buckets",
    "admin_project_usage_buckets",
    "admin_user_usage_buckets",
    # Resource Group Scoped Resolvers
    "rg_domain_usage_buckets",
    "rg_project_usage_buckets",
    "rg_user_usage_buckets",
    # Legacy Resolvers (deprecated)
    "domain_usage_buckets",
    "project_usage_buckets",
    "user_usage_buckets",
    # Common Types
    "UsageBucketMetadataGQL",
    # Domain Types
    "DomainUsageBucketGQL",
    "DomainUsageBucketConnection",
    "DomainUsageBucketEdge",
    "DomainUsageBucketFilter",
    "DomainUsageBucketOrderBy",
    # Project Types
    "ProjectUsageBucketGQL",
    "ProjectUsageBucketConnection",
    "ProjectUsageBucketEdge",
    "ProjectUsageBucketFilter",
    "ProjectUsageBucketOrderBy",
    # User Types
    "UserUsageBucketGQL",
    "UserUsageBucketConnection",
    "UserUsageBucketEdge",
    "UserUsageBucketFilter",
    "UserUsageBucketOrderBy",
    # Enums
    "UsageBucketOrderField",
)



"""
Adapters to convert scheduling history DTOs to repository BatchQuerier objects.
Handles conversion of filter, order, and pagination parameters.
Also provides data-to-DTO conversion functions.
"""

from __future__ import annotations

from ai.backend.common.dto.manager.scheduling_history import (
    DeploymentHistoryDTO,
    DeploymentHistoryFilter,
    DeploymentHistoryOrder,
    DeploymentHistoryOrderField,
    OrderDirection,
    RouteHistoryDTO,
    RouteHistoryFilter,
    RouteHistoryOrder,
    RouteHistoryOrderField,
    SchedulingResultType,
    SearchDeploymentHistoryRequest,
    SearchRouteHistoryRequest,
    SearchSessionHistoryRequest,
    SessionHistoryDTO,
    SessionHistoryFilter,
    SessionHistoryOrder,
    SessionHistoryOrderField,
    SubStepResultDTO,
)
from ai.backend.manager.api.adapter import BaseFilterAdapter
from ai.backend.manager.data.deployment.types import DeploymentHistoryData, RouteHistoryData
from ai.backend.manager.data.session.types import (
    SchedulingResult,
    SessionSchedulingHistoryData,
    SubStepResult,
)
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    OffsetPagination,
    QueryCondition,
    QueryOrder,
)
from ai.backend.manager.repositories.scheduling_history.options import (
    DeploymentHistoryConditions,
    DeploymentHistoryOrders,
    RouteHistoryConditions,
    RouteHistoryOrders,
    SessionSchedulingHistoryConditions,
    SessionSchedulingHistoryOrders,
)

__all__ = ("SchedulingHistoryAdapter",)


class SchedulingHistoryAdapter(BaseFilterAdapter):
    """Adapter for converting scheduling history requests to repository queries."""

    # Session History

    def build_session_history_querier(self, request: SearchSessionHistoryRequest) -> BatchQuerier:
        """Build a BatchQuerier for session scheduling history from search request."""
        conditions = self._convert_session_filter(request.filter) if request.filter else []
        orders = [self._convert_session_order(o) for o in request.order] if request.order else []
        pagination = OffsetPagination(limit=request.limit, offset=request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_session_filter(self, filter: SessionHistoryFilter) -> list[QueryCondition]:
        """Convert session history filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.session_id is not None:
            condition = self.convert_uuid_filter(
                filter.session_id,
                equals_factory=SessionSchedulingHistoryConditions.by_session_id_filter,
                in_factory=SessionSchedulingHistoryConditions.by_session_id_in,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.phase is not None:
            condition = self.convert_string_filter(
                filter.phase,
                contains_factory=SessionSchedulingHistoryConditions.by_phase_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_phase_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_phase_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_phase_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.from_status is not None and len(filter.from_status) > 0:
            conditions.append(
                SessionSchedulingHistoryConditions.by_from_statuses(filter.from_status)
            )

        if filter.to_status is not None and len(filter.to_status) > 0:
            conditions.append(SessionSchedulingHistoryConditions.by_to_statuses(filter.to_status))

        if filter.result is not None and len(filter.result) > 0:
            conditions.append(
                SessionSchedulingHistoryConditions.by_results([
                    self._convert_result_type(r) for r in filter.result
                ])
            )

        if filter.error_code is not None:
            condition = self.convert_string_filter(
                filter.error_code,
                contains_factory=SessionSchedulingHistoryConditions.by_error_code_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_error_code_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_error_code_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_error_code_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.message is not None:
            condition = self.convert_string_filter(
                filter.message,
                contains_factory=SessionSchedulingHistoryConditions.by_message_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_message_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_message_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_message_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions

    def _convert_session_order(self, order: SessionHistoryOrder) -> QueryOrder:
        """Convert session history order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        match order.field:
            case SessionHistoryOrderField.CREATED_AT:
                return SessionSchedulingHistoryOrders.created_at(ascending=ascending)
            case SessionHistoryOrderField.UPDATED_AT:
                return SessionSchedulingHistoryOrders.updated_at(ascending=ascending)

        raise ValueError(f"Unknown order field: {order.field}")

    def convert_session_history_to_dto(
        self, data: SessionSchedulingHistoryData
    ) -> SessionHistoryDTO:
        """Convert SessionSchedulingHistoryData to DTO."""
        return SessionHistoryDTO(
            id=data.id,
            session_id=data.session_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=data.result.value,
            error_code=data.error_code,
            message=data.message,
            sub_steps=[self._convert_sub_step(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )

    # Deployment History

    def build_deployment_history_querier(
        self, request: SearchDeploymentHistoryRequest
    ) -> BatchQuerier:
        """Build a BatchQuerier for deployment history from search request."""
        conditions = self._convert_deployment_filter(request.filter) if request.filter else []
        orders = [self._convert_deployment_order(o) for o in request.order] if request.order else []
        pagination = OffsetPagination(limit=request.limit, offset=request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_deployment_filter(self, filter: DeploymentHistoryFilter) -> list[QueryCondition]:
        """Convert deployment history filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.deployment_id is not None:
            condition = self.convert_uuid_filter(
                filter.deployment_id,
                equals_factory=DeploymentHistoryConditions.by_deployment_id_filter,
                in_factory=DeploymentHistoryConditions.by_deployment_id_in,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.phase is not None:
            condition = self.convert_string_filter(
                filter.phase,
                contains_factory=DeploymentHistoryConditions.by_phase_contains,
                equals_factory=DeploymentHistoryConditions.by_phase_equals,
                starts_with_factory=DeploymentHistoryConditions.by_phase_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_phase_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.from_status is not None and len(filter.from_status) > 0:
            conditions.append(DeploymentHistoryConditions.by_from_statuses(filter.from_status))

        if filter.to_status is not None and len(filter.to_status) > 0:
            conditions.append(DeploymentHistoryConditions.by_to_statuses(filter.to_status))

        if filter.result is not None and len(filter.result) > 0:
            conditions.append(
                DeploymentHistoryConditions.by_results([
                    self._convert_result_type(r) for r in filter.result
                ])
            )

        if filter.error_code is not None:
            condition = self.convert_string_filter(
                filter.error_code,
                contains_factory=DeploymentHistoryConditions.by_error_code_contains,
                equals_factory=DeploymentHistoryConditions.by_error_code_equals,
                starts_with_factory=DeploymentHistoryConditions.by_error_code_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_error_code_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.message is not None:
            condition = self.convert_string_filter(
                filter.message,
                contains_factory=DeploymentHistoryConditions.by_message_contains,
                equals_factory=DeploymentHistoryConditions.by_message_equals,
                starts_with_factory=DeploymentHistoryConditions.by_message_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_message_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions

    def _convert_deployment_order(self, order: DeploymentHistoryOrder) -> QueryOrder:
        """Convert deployment history order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        match order.field:
            case DeploymentHistoryOrderField.CREATED_AT:
                return DeploymentHistoryOrders.created_at(ascending=ascending)
            case DeploymentHistoryOrderField.UPDATED_AT:
                return DeploymentHistoryOrders.updated_at(ascending=ascending)

        raise ValueError(f"Unknown order field: {order.field}")

    def convert_deployment_history_to_dto(
        self, data: DeploymentHistoryData
    ) -> DeploymentHistoryDTO:
        """Convert DeploymentHistoryData to DTO."""
        return DeploymentHistoryDTO(
            id=data.id,
            deployment_id=data.deployment_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=data.result.value,
            error_code=data.error_code,
            message=data.message,
            sub_steps=[self._convert_sub_step(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )

    # Route History

    def build_route_history_querier(self, request: SearchRouteHistoryRequest) -> BatchQuerier:
        """Build a BatchQuerier for route history from search request."""
        conditions = self._convert_route_filter(request.filter) if request.filter else []
        orders = [self._convert_route_order(o) for o in request.order] if request.order else []
        pagination = OffsetPagination(limit=request.limit, offset=request.offset)

        return BatchQuerier(conditions=conditions, orders=orders, pagination=pagination)

    def _convert_route_filter(self, filter: RouteHistoryFilter) -> list[QueryCondition]:
        """Convert route history filter to list of query conditions."""
        conditions: list[QueryCondition] = []

        if filter.route_id is not None:
            condition = self.convert_uuid_filter(
                filter.route_id,
                equals_factory=RouteHistoryConditions.by_route_id_filter,
                in_factory=RouteHistoryConditions.by_route_id_in,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.deployment_id is not None:
            condition = self.convert_uuid_filter(
                filter.deployment_id,
                equals_factory=RouteHistoryConditions.by_deployment_id_filter,
                in_factory=RouteHistoryConditions.by_deployment_id_in,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.phase is not None:
            condition = self.convert_string_filter(
                filter.phase,
                contains_factory=RouteHistoryConditions.by_phase_contains,
                equals_factory=RouteHistoryConditions.by_phase_equals,
                starts_with_factory=RouteHistoryConditions.by_phase_starts_with,
                ends_with_factory=RouteHistoryConditions.by_phase_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.from_status is not None and len(filter.from_status) > 0:
            conditions.append(RouteHistoryConditions.by_from_statuses(filter.from_status))

        if filter.to_status is not None and len(filter.to_status) > 0:
            conditions.append(RouteHistoryConditions.by_to_statuses(filter.to_status))

        if filter.result is not None and len(filter.result) > 0:
            conditions.append(
                RouteHistoryConditions.by_results([
                    self._convert_result_type(r) for r in filter.result
                ])
            )

        if filter.error_code is not None:
            condition = self.convert_string_filter(
                filter.error_code,
                contains_factory=RouteHistoryConditions.by_error_code_contains,
                equals_factory=RouteHistoryConditions.by_error_code_equals,
                starts_with_factory=RouteHistoryConditions.by_error_code_starts_with,
                ends_with_factory=RouteHistoryConditions.by_error_code_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        if filter.message is not None:
            condition = self.convert_string_filter(
                filter.message,
                contains_factory=RouteHistoryConditions.by_message_contains,
                equals_factory=RouteHistoryConditions.by_message_equals,
                starts_with_factory=RouteHistoryConditions.by_message_starts_with,
                ends_with_factory=RouteHistoryConditions.by_message_ends_with,
            )
            if condition is not None:
                conditions.append(condition)

        return conditions

    def _convert_route_order(self, order: RouteHistoryOrder) -> QueryOrder:
        """Convert route history order specification to query order."""
        ascending = order.direction == OrderDirection.ASC

        match order.field:
            case RouteHistoryOrderField.CREATED_AT:
                return RouteHistoryOrders.created_at(ascending=ascending)
            case RouteHistoryOrderField.UPDATED_AT:
                return RouteHistoryOrders.updated_at(ascending=ascending)

        raise ValueError(f"Unknown order field: {order.field}")

    def convert_route_history_to_dto(self, data: RouteHistoryData) -> RouteHistoryDTO:
        """Convert RouteHistoryData to DTO."""
        return RouteHistoryDTO(
            id=data.id,
            route_id=data.route_id,
            deployment_id=data.deployment_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=data.result.value,
            error_code=data.error_code,
            message=data.message,
            sub_steps=[self._convert_sub_step(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )

    # Common helpers

    def _convert_result_type(self, result: SchedulingResultType) -> SchedulingResult:
        """Convert DTO result type to domain result type."""
        match result:
            case SchedulingResultType.SUCCESS:
                return SchedulingResult.SUCCESS
            case SchedulingResultType.FAILURE:
                return SchedulingResult.FAILURE
            case SchedulingResultType.STALE:
                return SchedulingResult.STALE

        raise ValueError(f"Unknown result type: {result}")

    def _convert_sub_step(self, sub_step: SubStepResult) -> SubStepResultDTO:
        """Convert SubStepResult to DTO."""
        return SubStepResultDTO(
            step=sub_step.step,
            result=sub_step.result.value,
            error_code=sub_step.error_code,
            message=sub_step.message,
            started_at=sub_step.started_at,
            ended_at=sub_step.ended_at,
        )



"""Action for creating deployments."""

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.creator import NewDeploymentCreator
from ai.backend.manager.data.deployment.types import ModelDeploymentData
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class CreateDeploymentAction(DeploymentBaseAction):
    """Action to create a new deployment(Model Service)."""

    creator: NewDeploymentCreator

    @override
    def entity_id(self) -> str | None:
        return None  # New deployment doesn't have an ID yet

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateDeploymentActionResult(BaseActionResult):
    data: ModelDeploymentData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact.types import ArtifactRevisionData
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class GetArtifactRevisionsAction(ArtifactAction):
    artifact_id: uuid.UUID

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ARTIFACT_REVISION

    @override
    def entity_id(self) -> str | None:
        return str(self.artifact_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class GetArtifactRevisionsActionResult(BaseActionResult):
    revisions: list[ArtifactRevisionData]

    @override
    def entity_id(self) -> str | None:
        return None



"""remove_permission_groups_table_and_fk

Revision ID: f41bbe0c0f12
Revises: 299deadfb77e
Create Date: 2026-02-11 02:43:17.099347

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "f41bbe0c0f12"
down_revision = "299deadfb77e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Drop FK constraints
    op.drop_constraint(
        op.f("fk_permissions_permission_group_id_permission_groups"),
        "permissions",
        type_="foreignkey",
    )
    op.execute(
        "ALTER TABLE object_permissions"
        " DROP CONSTRAINT IF EXISTS fk_object_permissions_permission_group_id_permission_groups"
    )

    # Drop indexes on permission_group_id columns
    op.drop_index("ix_id_permission_group_id", table_name="permissions")
    op.execute("DROP INDEX IF EXISTS ix_object_permissions_permission_group_id")

    # Drop permission_group_id columns
    op.drop_column("permissions", "permission_group_id")
    op.execute("ALTER TABLE object_permissions DROP COLUMN IF EXISTS permission_group_id")

    # Drop indexes and constraints on permission_groups table
    op.drop_index("ix_id_role_id_scope_id", table_name="permission_groups")
    op.drop_constraint("uq_permission_groups_role_scope", "permission_groups", type_="unique")

    # Drop permission_groups table
    op.drop_table("permission_groups")


def downgrade() -> None:
    conn = op.get_bind()

    # Recreate permission_groups table
    op.create_table(
        "permission_groups",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("scope_type", sa.VARCHAR(length=32), nullable=False),
        sa.Column("scope_id", sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_permission_groups")),
    )
    op.create_index(
        "ix_id_role_id_scope_id", "permission_groups", ["id", "role_id", "scope_id"], unique=False
    )
    op.create_unique_constraint(
        "uq_permission_groups_role_scope",
        "permission_groups",
        ["role_id", "scope_type", "scope_id"],
    )

    # Populate permission_groups from existing permissions data
    conn.execute(
        sa.text("""
        INSERT INTO permission_groups (role_id, scope_type, scope_id)
        SELECT DISTINCT role_id, scope_type, scope_id FROM permissions
    """)
    )

    # Add permission_group_id columns as nullable first
    op.add_column(
        "permissions",
        sa.Column("permission_group_id", GUID(), nullable=True),
    )
    op.add_column(
        "object_permissions",
        sa.Column("permission_group_id", GUID(), nullable=True),
    )

    # Backfill permissions.permission_group_id
    conn.execute(
        sa.text("""
        UPDATE permissions p
        SET permission_group_id = pg.id
        FROM permission_groups pg
        WHERE pg.role_id = p.role_id
          AND pg.scope_type = p.scope_type
          AND pg.scope_id = p.scope_id
    """)
    )

    # Backfill object_permissions.permission_group_id
    conn.execute(
        sa.text("""
        UPDATE object_permissions op
        SET permission_group_id = pg.id
        FROM permission_groups pg, association_scopes_entities ase
        WHERE pg.scope_type = ase.scope_type
          AND pg.scope_id = ase.scope_id
          AND ase.entity_type = op.entity_type
          AND ase.entity_id = op.entity_id
          AND pg.role_id = op.role_id
    """)
    )

    # Delete orphan object_permissions that couldn't be mapped to a permission_group
    conn.execute(
        sa.text("""
        DELETE FROM object_permissions WHERE permission_group_id IS NULL
    """)
    )

    # Set columns to NOT NULL after backfill
    op.alter_column("permissions", "permission_group_id", nullable=False)
    op.alter_column("object_permissions", "permission_group_id", nullable=False)

    # Restore indexes
    op.create_index(
        "ix_id_permission_group_id", "permissions", ["id", "permission_group_id"], unique=False
    )
    op.create_index(
        "ix_object_permissions_permission_group_id",
        "object_permissions",
        ["permission_group_id"],
        unique=False,
    )

    # Restore FK constraints
    op.create_foreign_key(
        op.f("fk_permissions_permission_group_id_permission_groups"),
        "permissions",
        "permission_groups",
        ["permission_group_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        op.f("fk_object_permissions_permission_group_id_permission_groups"),
        "object_permissions",
        "permission_groups",
        ["permission_group_id"],
        ["id"],
        ondelete="CASCADE",
    )



"""redesign rbac tables

Revision ID: 28fecac94e67
Revises: 643deb439458
Create Date: 2025-08-06 10:22:55.140194

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "28fecac94e67"
down_revision = "643deb439458"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.drop_table("object_permissions")


def downgrade() -> None:
    op.create_table(
        "object_permissions",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("status", sa.String(length=64), nullable=False, server_default="active"),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("entity_type", sa.String(length=32), nullable=False),
        sa.Column("entity_id", sa.String(length=64), nullable=False),
        sa.Column("operation", sa.String(length=32), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_object_permissions")),
    )
    # Restore index and constraint from 643deb439458 state
    op.create_index(
        "ix_role_id_entity_id",
        "object_permissions",
        ["status", "role_id", "entity_id"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_object_permissions_entity_id_operation",
        "object_permissions",
        ["entity_id", "operation"],
    )



"""Deployment controller for managing model services and deployments."""

import logging
import uuid
from dataclasses import dataclass

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.data.permission.types import EntityType, ScopeType
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.creator import DeploymentCreationDraft
from ai.backend.manager.data.deployment.scale import AutoScalingRule, AutoScalingRuleCreator
from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    RouteInfo,
    RouteSearchResult,
    RouteTrafficStatus,
)
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.models.deployment_policy import DeploymentPolicyData
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.base.rbac.entity_creator import RBACEntityCreator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.deployment import DeploymentRepository
from ai.backend.manager.repositories.deployment.creators.endpoint import LegacyEndpointCreatorSpec
from ai.backend.manager.repositories.deployment.options import RouteConditions
from ai.backend.manager.repositories.deployment.updaters import (
    DeploymentPolicyUpdaterSpec,
    DeploymentUpdaterSpec,
    RouteUpdaterSpec,
)
from ai.backend.manager.sokovan.deployment.revision_generator.registry import (
    RevisionGeneratorRegistry,
)
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.sokovan.scheduling_controller.types import SessionValidationSpec
from ai.backend.manager.types import OptionalState

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class DeploymentControllerArgs:
    """Arguments for initializing DeploymentController."""

    scheduling_controller: SchedulingController
    deployment_repository: DeploymentRepository
    config_provider: ManagerConfigProvider
    storage_manager: StorageSessionManager
    event_producer: EventProducer
    valkey_schedule: ValkeyScheduleClient
    revision_generator_registry: RevisionGeneratorRegistry


class DeploymentController:
    """Controller for deployment and model service management."""

    _scheduling_controller: SchedulingController
    _deployment_repository: DeploymentRepository
    _config_provider: ManagerConfigProvider
    _storage_manager: StorageSessionManager
    _event_producer: EventProducer
    _valkey_schedule: ValkeyScheduleClient
    _revision_generator_registry: RevisionGeneratorRegistry

    def __init__(self, args: DeploymentControllerArgs) -> None:
        """Initialize the deployment controller with required services."""
        self._scheduling_controller = args.scheduling_controller
        self._deployment_repository = args.deployment_repository
        self._config_provider = args.config_provider
        self._storage_manager = args.storage_manager
        self._event_producer = args.event_producer
        self._valkey_schedule = args.valkey_schedule
        self._revision_generator_registry = args.revision_generator_registry

    async def create_deployment(
        self,
        draft: DeploymentCreationDraft,
    ) -> DeploymentInfo:
        """
        Create a new deployment based on the provided specification.

        Args:
            draft: Deployment creation specification

        Returns:
            DeploymentInfo: Information about the created deployment
        """
        log.info("Creating deployment '{}' in project {}", draft.name, draft.project)

        # Pre-fetch default architecture from scaling group
        default_architecture = (
            await self._deployment_repository.get_default_architecture_from_scaling_group(
                draft.metadata.resource_group
            )
        )

        generator = self._revision_generator_registry.get(
            draft.draft_model_revision.execution.runtime_variant
        )
        model_revision = await generator.generate_revision(
            draft_revision=draft.draft_model_revision,
            vfolder_id=draft.draft_model_revision.mounts.model_vfolder_id,
            default_architecture=default_architecture,
        )
        await self._scheduling_controller.validate_session_spec(
            SessionValidationSpec.from_revision(model_revision=model_revision)
        )
        image_id = await self._deployment_repository.get_image_id(model_revision.image_identifier)

        spec = LegacyEndpointCreatorSpec.from_deployment_creator(
            creator=draft.to_creator(model_revision),
            image_id=image_id,
        )
        creator = RBACEntityCreator(
            spec=spec,
            scope_ref=ScopeId(scope_type=ScopeType.USER, scope_id=str(draft.metadata.created_user)),
            additional_scope_refs=[],
            entity_type=EntityType.MODEL_DEPLOYMENT,
        )
        return await self._deployment_repository.create_endpoint_legacy(creator)

    async def update_deployment(
        self,
        endpoint_id: uuid.UUID,
        spec: DeploymentUpdaterSpec,
    ) -> DeploymentInfo:
        """
        Update an existing deployment with new specifications.

        Args:
            endpoint_id: ID of the deployment to update
            spec: Deployment updater specification

        Returns:
            DeploymentInfo: Information about the updated deployment
        """
        log.info("Updating deployment {}", endpoint_id)
        updater = Updater[EndpointRow](spec=spec, pk_value=endpoint_id)
        modified_endpoint = await self._deployment_repository.get_modified_endpoint(
            endpoint_id=endpoint_id, updater=updater
        )
        target_revision = modified_endpoint.target_revision()
        if target_revision:
            await self._scheduling_controller.validate_session_spec(
                SessionValidationSpec.from_revision(model_revision=target_revision)
            )
        res = await self._deployment_repository.update_endpoint_with_spec(updater)
        try:
            await self.mark_lifecycle_needed(DeploymentLifecycleType.CHECK_REPLICA)
        except Exception as e:
            log.error("Failed to mark deployment lifecycle needed: {}", e)
        return res

    async def destroy_deployment(
        self,
        endpoint_id: uuid.UUID,
    ) -> bool:
        """
        Destroy an existing deployment and its associated model service.

        Args:
            endpoint_id: ID of the endpoint to terminate
        Returns:
            bool: True if termination was successful, False otherwise
        """
        return await self._deployment_repository.destroy_endpoint(endpoint_id)

    async def create_autoscaling_rule(
        self,
        endpoint_id: uuid.UUID,
        creator: AutoScalingRuleCreator,
    ) -> AutoScalingRule:
        return await self._deployment_repository.create_autoscaling_rule(
            endpoint_id,
            creator,
        )

    async def list_autoscaling_rules(
        self,
        endpoint_id: uuid.UUID,
    ) -> list[AutoScalingRule]:
        return await self._deployment_repository.list_autoscaling_rules(
            endpoint_id,
        )

    async def delete_autoscaling_rule(
        self,
        rule_id: uuid.UUID,
    ) -> bool:
        return await self._deployment_repository.delete_autoscaling_rule(
            rule_id,
        )

    async def mark_lifecycle_needed(self, lifecycle_type: DeploymentLifecycleType) -> None:
        """
        Mark that a deployment lifecycle operation is needed for the next cycle.

        This is the public interface for hinting that deployment lifecycle operations
        should be processed. The actual processing will be handled by the coordinator.

        Args:
            lifecycle_type: Type of deployment lifecycle to mark as needed
        """
        await self._valkey_schedule.mark_deployment_needed(lifecycle_type.value)
        log.debug("Marked deployment lifecycle needed for type: {}", lifecycle_type.value)

    # ========== Deployment Policy Methods ==========

    async def get_deployment_policy(
        self,
        endpoint_id: uuid.UUID,
    ) -> DeploymentPolicyData:
        """Get the deployment policy for an endpoint.

        Args:
            endpoint_id: ID of the endpoint

        Returns:
            DeploymentPolicyData: Policy data
        """
        return await self._deployment_repository.get_deployment_policy(endpoint_id)

    async def update_deployment_policy(
        self,
        endpoint_id: uuid.UUID,
        updater_spec: DeploymentPolicyUpdaterSpec,
    ) -> DeploymentPolicyData:
        """Update the deployment policy for an endpoint.

        Args:
            endpoint_id: ID of the endpoint
            updater_spec: Policy update specification

        Returns:
            DeploymentPolicyData: Updated policy data
        """
        # First get the policy to find its ID (primary key)
        policy = await self._deployment_repository.get_deployment_policy(endpoint_id)
        return await self._deployment_repository.update_deployment_policy(
            Updater(spec=updater_spec, pk_value=policy.id)
        )

    # Route operations

    async def search_routes(
        self,
        querier: BatchQuerier,
    ) -> RouteSearchResult:
        """Search routes with pagination and filtering.

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination

        Returns:
            RouteSearchResult with items, total_count, and pagination info
        """
        return await self._deployment_repository.search_routes(querier)

    async def get_route(
        self,
        route_id: uuid.UUID,
    ) -> RouteInfo | None:
        """Get a single route by its ID.

        Args:
            route_id: The route ID

        Returns:
            RouteInfo if found, None otherwise
        """
        querier = BatchQuerier(
            pagination=OffsetPagination(limit=1),
            conditions=[RouteConditions.by_ids([route_id])],
        )
        result = await self._deployment_repository.search_routes(querier)
        return result.items[0] if result.items else None

    async def update_route_traffic_status(
        self,
        route_id: uuid.UUID,
        traffic_status: RouteTrafficStatus,
    ) -> RouteInfo | None:
        """Update route traffic status.

        Args:
            route_id: The route ID
            traffic_status: New traffic status

        Returns:
            Updated RouteInfo if found, None otherwise
        """
        spec = RouteUpdaterSpec(
            traffic_status=OptionalState.update(traffic_status),
        )
        updater: Updater[RoutingRow] = Updater(spec=spec, pk_value=route_id)
        success = await self._deployment_repository.update_route(updater)
        if not success:
            return None
        querier = BatchQuerier(
            pagination=OffsetPagination(limit=1),
            conditions=[RouteConditions.by_ids([route_id])],
        )
        result = await self._deployment_repository.search_routes(querier)
        return result.items[0] if result.items else None



"""GraphQL types for auto-scaling rules."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from typing import Any, Self, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.common.types import AutoScalingMetricSource as CommonAutoScalingMetricSource
from ai.backend.manager.api.gql.base import DateTimeFilter, OrderDirection
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.deployment.scale import ModelDeploymentAutoScalingRuleCreator
from ai.backend.manager.data.deployment.scale_modifier import ModelDeploymentAutoScalingRuleModifier
from ai.backend.manager.data.deployment.types import (
    AutoScalingRuleOrderField,
    ModelDeploymentAutoScalingRuleData,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.deployment.options import (
    AutoScalingRuleConditions,
    AutoScalingRuleOrders,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.update_auto_scaling_rule import (
    UpdateAutoScalingRuleAction,
)
from ai.backend.manager.types import OptionalState


@strawberry.enum(description="Added in 25.1.0")
class AutoScalingMetricSource(StrEnum):
    KERNEL = "KERNEL"
    INFERENCE_FRAMEWORK = "INFERENCE_FRAMEWORK"


@strawberry.input(description="Added in 25.19.0")
class AutoScalingRuleFilter(GQLFilter):
    """Filter for auto-scaling rules."""

    created_at: DateTimeFilter | None = None
    last_triggered_at: DateTimeFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.created_at:
            condition = self.created_at.build_query_condition(
                before_factory=AutoScalingRuleConditions.by_created_at_before,
                after_factory=AutoScalingRuleConditions.by_created_at_after,
                equals_factory=AutoScalingRuleConditions.by_created_at_equals,
            )
            if condition:
                conditions.append(condition)

        if self.last_triggered_at:
            condition = self.last_triggered_at.build_query_condition(
                before_factory=AutoScalingRuleConditions.by_last_triggered_at_before,
                after_factory=AutoScalingRuleConditions.by_last_triggered_at_after,
                equals_factory=AutoScalingRuleConditions.by_last_triggered_at_equals,
            )
            if condition:
                conditions.append(condition)

        return conditions


@strawberry.input(description="Added in 25.19.0")
class AutoScalingRuleOrderBy(GQLOrderBy):
    field: AutoScalingRuleOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case AutoScalingRuleOrderField.CREATED_AT:
                return AutoScalingRuleOrders.created_at(ascending)


@strawberry.type
class AutoScalingRule(Node):
    id: NodeID[str]

    metric_source: AutoScalingMetricSource = strawberry.field(
        description="Added in 25.19.0 (e.g. KERNEL, INFERENCE_FRAMEWORK)"
    )
    metric_name: str = strawberry.field()

    min_threshold: Decimal | None = strawberry.field(
        description="Added in 25.19.0: The minimum threshold for scaling (e.g. 0.5)"
    )
    max_threshold: Decimal | None = strawberry.field(
        description="Added in 25.19.0: The maximum threshold for scaling (e.g. 21.0)"
    )

    step_size: int = strawberry.field(
        description="Added in 25.19.0: The step size for scaling (e.g. 1)."
    )
    time_window: int = strawberry.field(
        description="Added in 25.19.0: The time window (seconds) for scaling (e.g. 60)."
    )

    min_replicas: int | None = strawberry.field(
        description="Added in 25.19.0: The minimum number of replicas (e.g. 1)."
    )
    max_replicas: int | None = strawberry.field(
        description="Added in 25.19.0: The maximum number of replicas (e.g. 10)."
    )

    created_at: datetime
    last_triggered_at: datetime

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.auto_scaling_rule_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ModelDeploymentAutoScalingRuleData) -> Self:
        return cls(
            id=ID(str(data.id)),
            metric_source=AutoScalingMetricSource(data.metric_source.name),
            metric_name=data.metric_name,
            min_threshold=data.min_threshold,
            max_threshold=data.max_threshold,
            step_size=data.step_size,
            time_window=data.time_window,
            min_replicas=data.min_replicas,
            max_replicas=data.max_replicas,
            created_at=data.created_at,
            last_triggered_at=data.last_triggered_at,
        )


AutoScalingRuleEdge = Edge[AutoScalingRule]


@strawberry.type(description="Added in 25.19.0")
class AutoScalingRuleConnection(Connection[AutoScalingRule]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


# Input Types
@strawberry.input
class CreateAutoScalingRuleInput:
    model_deployment_id: ID
    metric_source: AutoScalingMetricSource
    metric_name: str
    min_threshold: Decimal | None
    max_threshold: Decimal | None
    step_size: int
    time_window: int
    min_replicas: int | None
    max_replicas: int | None

    def to_creator(self) -> ModelDeploymentAutoScalingRuleCreator:
        return ModelDeploymentAutoScalingRuleCreator(
            model_deployment_id=UUID(self.model_deployment_id),
            metric_source=CommonAutoScalingMetricSource(self.metric_source.lower()),
            metric_name=self.metric_name,
            min_threshold=self.min_threshold,
            max_threshold=self.max_threshold,
            step_size=self.step_size,
            time_window=self.time_window,
            min_replicas=self.min_replicas,
            max_replicas=self.max_replicas,
        )


@strawberry.input
class UpdateAutoScalingRuleInput:
    id: ID
    metric_source: AutoScalingMetricSource | None
    metric_name: str | None
    min_threshold: Decimal | None
    max_threshold: Decimal | None
    step_size: int | None
    time_window: int | None
    min_replicas: int | None
    max_replicas: int | None

    def to_action(self) -> UpdateAutoScalingRuleAction:
        optional_state_metric_source = OptionalState[CommonAutoScalingMetricSource].nop()
        if isinstance(self.metric_source, AutoScalingMetricSource):
            optional_state_metric_source = OptionalState[CommonAutoScalingMetricSource].update(
                CommonAutoScalingMetricSource(self.metric_source)
            )
        return UpdateAutoScalingRuleAction(
            auto_scaling_rule_id=UUID(self.id),
            modifier=ModelDeploymentAutoScalingRuleModifier(
                metric_source=optional_state_metric_source,
                metric_name=OptionalState[str].from_graphql(self.metric_name),
                min_threshold=OptionalState[Decimal].from_graphql(self.min_threshold),
                max_threshold=OptionalState[Decimal].from_graphql(self.max_threshold),
                step_size=OptionalState[int].from_graphql(self.step_size),
                time_window=OptionalState[int].from_graphql(self.time_window),
                min_replicas=OptionalState[int].from_graphql(self.min_replicas),
                max_replicas=OptionalState[int].from_graphql(self.max_replicas),
            ),
        )


@strawberry.input
class DeleteAutoScalingRuleInput:
    id: ID


# Payload Types
@strawberry.type
class CreateAutoScalingRulePayload:
    auto_scaling_rule: AutoScalingRule


@strawberry.type
class UpdateAutoScalingRulePayload:
    auto_scaling_rule: AutoScalingRule


@strawberry.type
class DeleteAutoScalingRulePayload:
    id: ID
